from textwrap import dedent
from unittest import mock
from unittest.mock import Mock, patch

import pytest

from pyspark.sql import SparkSession
from pyspark.sql import types as t

from koheesio.spark.snowflake import (
    AddColumn,
    CreateOrReplaceTableFromDataFrame,
    DbTableQuery,
    GetTableSchema,
    GrantPrivilegesOnObject,
    GrantPrivilegesOnTable,
    GrantPrivilegesOnView,
    Query,
    RunQuery,
    SnowflakeBaseModel,
    SnowflakeReader,
    SnowflakeWriter,
    SyncTableAndDataFrameSchema,
    TableExists,
    TagSnowflakeQuery,
    map_spark_type,
)
from koheesio.spark.writers import BatchOutputMode

pytestmark = pytest.mark.spark

COMMON_OPTIONS = {
    "url": "url",
    "user": "user",
    "password": "password",
    "database": "db",
    "schema": "schema",
    "role": "role",
    "warehouse": "warehouse",
}


def test_snowflake_module_import():
    # test that the pass-through imports in the koheesio.spark snowflake modules are working
    from koheesio.spark.readers import snowflake as snowflake_writers
    from koheesio.spark.writers import snowflake as snowflake_readers


class TestSnowflakeReader:
    @pytest.mark.parametrize(
        "reader_options", [{"dbtable": "table", **COMMON_OPTIONS}, {"table": "table", **COMMON_OPTIONS}]
    )
    def test_get_options(self, reader_options):
        sf = SnowflakeReader(**(reader_options | {"authenticator": None}))
        o = sf.get_options()
        assert sf.format == "snowflake"
        assert o["sfUser"] == "user"
        assert o["sfCompress"] == "on"
        assert "authenticator" not in o

    @pytest.mark.parametrize(
        "reader_options", [{"dbtable": "table", **COMMON_OPTIONS}, {"table": "table", **COMMON_OPTIONS}]
    )
    def test_execute(self, dummy_spark, reader_options):
        """Method should be callable from parent class"""
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = SnowflakeReader(**reader_options).execute()
            assert k.df.count() == 1


class TestRunQuery:
    query_options = {"query": "query", **COMMON_OPTIONS}

    def test_get_options(self):
        k = RunQuery(**self.query_options)
        o = k.get_options()

        assert o["host"] == o["sfURL"]

    def test_execute(self, dummy_spark):
        pass


class TestQuery:
    query_options = {"query": "query", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = Query(**self.query_options)
            assert k.df.count() == 1


class TestTableQuery:
    options = {"table": "table", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = DbTableQuery(**self.options).execute()
            assert k.df.count() == 1


class TestTableExists:
    table_exists_options = {"table": "table", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = TableExists(**self.table_exists_options).execute()
            assert k.exists is True


class TestCreateOrReplaceTableFromDataFrame:
    options = {"table": "table", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark, dummy_df):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = CreateOrReplaceTableFromDataFrame(**self.options, df=dummy_df).execute()
            assert k.snowflake_schema == "id BIGINT"
            assert k.query == "CREATE OR REPLACE TABLE db.schema.table (id BIGINT)"
            assert len(k.input_schema) > 0


class TestGetTableSchema:
    get_table_schema_options = {"table": "table", **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = GetTableSchema(**self.get_table_schema_options)
            assert len(k.execute().table_schema.fields) == 1


class TestAddColumn:
    options = {"table": "foo", "column": "bar", "type": t.DateType(), **COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = AddColumn(**self.options).execute()
            assert k.query == "ALTER TABLE FOO ADD COLUMN BAR DATE"


def test_grant_privileges_on_object(dummy_spark):
    options = dict(
        **COMMON_OPTIONS, object="foo", type="TABLE", privileges=["DELETE", "SELECT"], roles=["role_1", "role_2"]
    )
    del options["role"]  # role is not required for this step as we are setting "roles"

    kls = GrantPrivilegesOnObject(**options)

    with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
        mock_spark.return_value = dummy_spark
        k = kls.execute()

        assert len(k.query) == 2, "expecting 2 queries (one for each role)"
        assert "DELETE" in k.query[0]
        assert "SELECT" in k.query[0]


def test_grant_privileges_on_table(dummy_spark):
    options = {**COMMON_OPTIONS, **dict(table="foo", privileges=["SELECT"], roles=["role_1"])}
    del options["role"]  # role is not required for this step as we are setting "roles"

    kls = GrantPrivilegesOnTable(
        **options,
    )
    with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
        mock_spark.return_value = dummy_spark

        k = kls.execute()
        assert k.query == [
            "GRANT SELECT ON TABLE DB.SCHEMA.FOO TO ROLE ROLE_1",
        ]


class TestGrantPrivilegesOnView:
    options = {**COMMON_OPTIONS}

    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = GrantPrivilegesOnView(**self.options, view="foo", privileges=["SELECT"], roles=["role_1"]).execute()
            assert k.query == [
                "GRANT SELECT ON VIEW DB.SCHEMA.FOO TO ROLE ROLE_1",
            ]


class TestSnowflakeWriter:
    def test_execute(self, dummy_spark):
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            k = SnowflakeWriter(
                **COMMON_OPTIONS,
                table="foo",
                df=dummy_spark.load(),
                mode=BatchOutputMode.OVERWRITE,
            )
            k.execute()


class TestSyncTableAndDataFrameSchema:
    @mock.patch("koheesio.spark.snowflake.AddColumn")
    @mock.patch("koheesio.spark.snowflake.GetTableSchema")
    def test_execute(self, mock_get_table_schema, mock_add_column, spark, caplog):
        from pyspark.sql.types import StringType, StructField, StructType

        df = spark.createDataFrame(data=[["val"]], schema=["foo"])
        sf_schema_before = StructType([StructField("bar", StringType(), True)])
        sf_schema_after = StructType([StructField("bar", StringType(), True), StructField("foo", StringType(), True)])

        mock_get_table_schema_instance = mock_get_table_schema()
        mock_get_table_schema_instance.execute.side_effect = [
            mock.Mock(table_schema=sf_schema_before),
            mock.Mock(table_schema=sf_schema_after),
        ]

        with caplog.at_level("DEBUG"):
            k = SyncTableAndDataFrameSchema(
                **COMMON_OPTIONS,
                table="foo",
                df=df,
                dry_run=True,
            ).execute()
            print(f"{caplog.text = }")
            assert "Columns to be added to Snowflake table: {'foo'}" in caplog.text
            assert "Columns to be added to Spark DataFrame: {'bar'}" in caplog.text
            assert k.new_df_schema == StructType()

        k = SyncTableAndDataFrameSchema(
            **COMMON_OPTIONS,
            table="foo",
            df=df,
        ).execute()
        assert k.df.columns == ["bar", "foo"]


@pytest.mark.parametrize(
    "input_value,expected",
    [
        (t.BinaryType(), "VARBINARY"),
        (t.BooleanType(), "BOOLEAN"),
        (t.ByteType(), "BINARY"),
        (t.DateType(), "DATE"),
        (t.TimestampType(), "TIMESTAMP"),
        (t.DoubleType(), "DOUBLE"),
        (t.FloatType(), "FLOAT"),
        (t.IntegerType(), "INT"),
        (t.LongType(), "BIGINT"),
        (t.NullType(), "STRING"),
        (t.ShortType(), "SMALLINT"),
        (t.StringType(), "STRING"),
        (t.NumericType(), "FLOAT"),
        (t.DecimalType(0, 1), "DECIMAL(0,1)"),
        (t.DecimalType(0, 100), "DECIMAL(0,100)"),
        (t.DecimalType(10, 0), "DECIMAL(10,0)"),
        (t.DecimalType(), "DECIMAL(10,0)"),
        (t.MapType(t.IntegerType(), t.StringType()), "VARIANT"),
        (t.ArrayType(t.StringType()), "VARIANT"),
        (t.StructType([t.StructField(name="foo", dataType=t.StringType())]), "VARIANT"),
        (t.DayTimeIntervalType(), "STRING"),
    ],
)
def test_map_spark_type(input_value, expected):
    assert map_spark_type(input_value) == expected


class TestSnowflakeBaseModel:
    def test_get_options(self, dummy_spark):
        k = SnowflakeBaseModel(
            sfURL="url",
            sfUser="user",
            sfPassword="password",
            sfDatabase="database",
            sfRole="role",
            sfWarehouse="warehouse",
            schema="schema",
        )
        options = k.get_options()
        assert options["sfURL"] == "url"
        assert options["sfUser"] == "user"
        assert options["sfDatabase"] == "database"
        assert options["sfRole"] == "role"
        assert options["sfWarehouse"] == "warehouse"
        assert options["sfSchema"] == "schema"


class TestTagSnowflakeQuery:
    def test_tag_query_no_existing_preactions(self):
        expected_preactions = (
            """ALTER SESSION SET QUERY_TAG = '{"pipeline_name": "test-pipeline-1","task_name": "test_task_1"}';"""
        )

        tagged_options = (
            TagSnowflakeQuery(
                task_name="test_task_1",
                pipeline_name="test-pipeline-1",
            )
            .execute()
            .options
        )

        assert len(tagged_options) == 1
        preactions = tagged_options["preactions"].replace("    ", "").replace("\n", "")
        assert preactions == expected_preactions

    def test_tag_query_present_existing_preactions(self):
        options = {
            "otherSfOption": "value",
            "preactions": "SET TEST_VAR = 'ABC';",
        }
        query_tag_preaction = (
            """ALTER SESSION SET QUERY_TAG = '{"pipeline_name": "test-pipeline-2","task_name": "test_task_2"}';"""
        )
        expected_preactions = f"SET TEST_VAR = 'ABC';{query_tag_preaction}" ""

        tagged_options = (
            TagSnowflakeQuery(task_name="test_task_2", pipeline_name="test-pipeline-2", options=options)
            .execute()
            .options
        )

        assert len(tagged_options) == 2
        assert tagged_options["otherSfOption"] == "value"
        preactions = tagged_options["preactions"].replace("    ", "").replace("\n", "")
        assert preactions == expected_preactions


def test_table_exists(spark):
    # Create a TableExists instance
    te = TableExists(
        sfURL="url",
        sfUser="user",
        sfPassword="password",
        sfDatabase="database",
        sfRole="role",
        sfWarehouse="warehouse",
        schema="schema",
        table="table",
    )

    expected_query = dedent(
        """
        SELECT *
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_CATALOG     = 'DATABASE'
          AND TABLE_SCHEMA      = 'SCHEMA'
          AND TABLE_TYPE        = 'BASE TABLE'
          AND UPPER(TABLE_NAME) = 'TABLE'
        """
    ).strip()

    # Create a Mock object for the Query class
    mock_query = Mock(spec=Query)
    mock_query.read.return_value = spark.range(1)

    # Patch the Query class to return the mock_query when instantiated
    with patch("koheesio.spark.snowflake.Query", return_value=mock_query) as mock_query_class:
        # Execute the SnowflakeBaseModel instance
        te.execute()

        # Assert that the query is as expected
        assert mock_query_class.call_args[1]["query"] == expected_query



"""
This module contains the base classes for all transformations.

See class docstrings for more information.

References
----------
For a comprehensive guide on the usage, examples, and additional features of Transformation classes, please refer to the
[reference/concepts/spark/transformations](../../../reference/spark/transformations.md) section of the Koheesio
documentation.

Classes
-------
Transformation
    Base class for all transformations

ColumnsTransformation
    Extended Transformation class with a preset validator for handling column(s) data

ColumnsTransformationWithTarget
    Extended ColumnsTransformation class with an additional `target_column` field
"""

from typing import List, Optional, Union
from abc import ABC, abstractmethod

from pyspark.sql import Column
from pyspark.sql import functions as f
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import DataType

from koheesio.models import Field, ListOfColumns, field_validator
from koheesio.spark import SparkStep
from koheesio.spark.utils import SparkDatatype


class Transformation(SparkStep, ABC):
    """Base class for all transformations

    Concept
    -------
    A Transformation is a Step that takes a DataFrame as input and returns a DataFrame as output. The DataFrame is
    transformed based on the logic implemented in the `execute` method. Any additional parameters that are needed for
    the transformation can be passed to the constructor.

    Parameters
    ----------
    df : Optional[DataFrame]
        The DataFrame to apply the transformation to. If not provided, the DataFrame has to be passed to the
        transform-method.

    Example
    -------
    ```python
    from koheesio.steps.transformations import Transformation
    from pyspark.sql import functions as f


    class AddOne(Transformation):
        def execute(self):
            self.output.df = self.df.withColumn(
                "new_column", f.col("old_column") + 1
            )
    ```

    In the example above, the `execute` method is implemented to add 1 to the values of the `old_column` and store the
    result in a new column called `new_column`.

    In order to use this transformation, we can call the `transform` method:

    ```python
    from pyspark.sql import SparkSession

    # create a DataFrame with 3 rows
    df = SparkSession.builder.getOrCreate().range(3)

    output_df = AddOne().transform(df)
    ```

    The `output_df` will now contain the original DataFrame with an additional column called `new_column` with the
    values of `old_column` + 1.

    __output_df:__

    |id|new_column|
    |--|----------|
    | 0|         1|
    | 1|         2|
    | 2|         3|
    ...

    Alternatively, we can pass the DataFrame to the constructor and call the `execute` or `transform` method without
    any arguments:

    ```python
    output_df = AddOne(df).transform()
    # or
    output_df = AddOne(df).execute().output.df
    ```

    Note: that the transform method was not implemented explicitly in the AddOne class. This is because the `transform`
    method is already implemented in the `Transformation` class. This means that all classes that inherit from the
    Transformation class will have the `transform` method available. Only the execute method needs to be implemented.
    """

    df: Optional[DataFrame] = Field(default=None, description="The Spark DataFrame")

    @abstractmethod
    def execute(self) -> SparkStep.Output:
        """Execute on a Transformation should handle self.df (input) and set self.output.df (output)

        This method should be implemented in the child class. The input DataFrame is available as `self.df` and the
        output DataFrame should be stored in `self.output.df`.

        For example:
        ```python
        def execute(self):
            self.output.df = self.df.withColumn("new_column", f.col("old_column") + 1)
        ```

        The transform method will call this method and return the output DataFrame.
        """
        # self.df  # input dataframe
        # self.output.df # output dataframe
        self.output.df = ...  # implement the transformation logic
        raise NotImplementedError

    def transform(self, df: Optional[DataFrame] = None) -> DataFrame:
        """Execute the transformation and return the output DataFrame

        Note: when creating a child from this, don't implement this transform method. Instead, implement execute!

        See Also
        --------
        `Transformation.execute`

        Parameters
        ----------
        df: Optional[DataFrame]
            The DataFrame to apply the transformation to. If not provided, the DataFrame passed to the constructor
            will be used.

        Returns
        -------
        DataFrame
            The transformed DataFrame
        """
        self.df = df or self.df
        if not self.df:
            raise RuntimeError("No valid Dataframe was passed")
        self.execute()
        return self.output.df


class ColumnsTransformation(Transformation, ABC):
    """Extended Transformation class with a preset validator for handling column(s) data with a standardized input
    for a single column or multiple columns.

    Concept
    -------
    A ColumnsTransformation is a Transformation with a standardized input for column or columns.

    - `columns` are stored as a list
    - either a single string, or a list of strings can be passed to enter the `columns`
    - `column` and `columns` are aliases to one another - internally the name `columns` should be used though.

    If more than one column is passed, the behavior of the Class changes this way:

    - the transformation will be run in a loop against all the given columns

    Configuring the ColumnsTransformation
    -------------------------------------
    [ColumnConfig]: ./index.md#koheesio.spark.transformations.ColumnsTransformation.ColumnConfig
    [SparkDatatype]: ../utils.md#koheesio.spark.utils.SparkDatatype

    The ColumnsTransformation class has a [ColumnConfig] class that can be used to configure the behavior of the class.
    Users should not have to interact with the [ColumnConfig] class directly.

    This class has the following fields:

    - `run_for_all_data_type`
        allows to run the transformation for all columns of a given type.

    - `limit_data_type`
        allows to limit the transformation to a specific data type.

    - `data_type_strict_mode`
        Toggles strict mode for data type validation. Will only work if `limit_data_type` is set.

    Data types need to be specified as a [SparkDatatype] enum.

    ---

    <small>
    - See the docstrings of the [ColumnConfig] class for more information.<br>
    - See the [SparkDatatype] enum for a list of available data types.<br>
    </small>

    Example
    -------
    Implementing a transformation using the `ColumnsTransformation` class:

    ```python
    from pyspark.sql import functions as f
    from koheesio.steps.transformations import ColumnsTransformation


    class AddOne(ColumnsTransformation):
        def execute(self):
            for column in self.get_columns():
                self.output.df = self.df.withColumn(column, f.col(column) + 1)
    ```

    In the above example, the `execute` method is implemented to add 1 to the values of a given column.

    Parameters
    ----------
    columns : ListOfColumns
        The column (or list of columns) to apply the transformation to. Alias: column

    """

    columns: ListOfColumns = Field(
        default="",
        alias="column",
        description="The column (or list of columns) to apply the transformation to. Alias: column",
    )

    class ColumnConfig:
        """
        Koheesio ColumnsTransformation specific Config

        Parameters
        ----------
        run_for_all_data_type : Optional[List[SparkDatatype]]
            allows to run the transformation for all columns of a given type.
            A user can trigger this behavior by either omitting the `columns` parameter or by passing a single `*` as a
            column name. In both cases, the `run_for_all_data_type` will be used to determine the data type.
            Value should be be passed as a SparkDatatype enum.
            (default: [None])

        limit_data_type : Optional[List[SparkDatatype]]
            allows to limit the transformation to a specific data type.
            Value should be passed as a SparkDatatype enum.
            (default: [None])

        data_type_strict_mode : bool
            Toggles strict mode for data type validation. Will only work if `limit_data_type` is set.
            - when True, a ValueError will be raised if any column does not adhere to the `limit_data_type`
            - when False, a warning will be thrown and the column will be skipped instead
            (default: False)
        """

        run_for_all_data_type: Optional[List[SparkDatatype]] = [None]
        limit_data_type: Optional[List[SparkDatatype]] = [None]
        data_type_strict_mode: bool = False

    @field_validator("columns", mode="before")
    def set_columns(cls, columns_value):
        """Validate columns through the columns configuration provided"""
        columns = columns_value
        run_for_all_data_type = cls.ColumnConfig.run_for_all_data_type

        if run_for_all_data_type and len(columns) == 0:
            columns = ["*"]

        if columns[0] == "*" and not run_for_all_data_type:
            raise ValueError("Cannot use '*' as a column name when no run_for_all_data_type is set")

        return columns

    @property
    def run_for_all_is_set(self) -> bool:
        """Returns True if the transformation should be run for all columns of a given type"""
        run_for_all_data_type = self.ColumnConfig.run_for_all_data_type
        if run_for_all_data_type and self.columns[0] == "*":
            return True
        return False

    @property
    def limit_data_type_is_set(self) -> bool:
        """Returns True if limit_data_type is set"""
        return self.ColumnConfig.limit_data_type[0] is not None

    @property
    def data_type_strict_mode_is_set(self) -> bool:
        """Returns True if data_type_strict_mode is set"""
        return self.ColumnConfig.data_type_strict_mode

    def column_type_of_col(
        self, col: Union[str, Column], df: Optional[DataFrame] = None, simple_return_mode: bool = True
    ) -> Union[DataType, str]:
        """
        Returns the dataType of a Column object as a string.

        The Column object does not have a type attribute, so we have to ask the DataFrame its schema and find the type
        based on the column name. We retrieve the name of the column from the Column object by calling toString() from
        the JVM.

        Examples
        --------
        __input_df:__
        | str_column | int_column |
        |------------|------------|
        | hello      | 1          |
        | world      | 2          |

        ```python
        # using the AddOne transformation from the example above
        add_one = AddOne(
            columns=["str_column", "int_column"],
            df=input_df,
        )
        add_one.column_type_of_col("str_column")  # returns "string"
        add_one.column_type_of_col("int_column")  # returns "integer"
        # returns IntegerType
        add_one.column_type_of_col("int_column", simple_return_mode=False)
        ```

        Parameters
        ----------
        col : Union[str, Column]
            The column to check the type of

        df : Optional[DataFrame]
            The DataFrame belonging to the column. If not provided, the DataFrame passed to the constructor
            will be used.

        simple_return_mode : bool
            If True, the return value will be a simple string. If False, the return value will be a SparkDatatype enum.

        Returns
        -------
        datatype: str
            The type of the column as a string
        """
        df = df or self.df
        if not df:
            raise RuntimeError("No valid Dataframe was passed")

        if not isinstance(col, Column):
            col = f.col(col)

        # ask the JVM for the name of the column
        # noinspection PyProtectedMember
        col_name = col._jc.toString()

        # In order to check the datatype of the column, we have to ask the DataFrame its schema
        df_col = [c for c in df.schema if c.name == col_name][0]

        if simple_return_mode:
            return SparkDatatype(df_col.dataType.typeName()).value

        return df_col.dataType

    def get_all_columns_of_specific_type(self, data_type: Union[str, SparkDatatype]) -> List[str]:
        """Get all columns from the dataframe of a given type

        A DataFrame needs to be available in order to get the columns. If no DataFrame is available, a ValueError will
        be raised.

        Note: only one data type can be passed to this method. If you want to get columns of multiple data types, you
        have to call this method multiple times.

        Parameters
        ----------
        data_type : Union[str, SparkDatatype]
            The data type to get the columns for

        Returns
        -------
        List[str]
            A list of column names of the given data type
        """
        if not self.df:
            raise ValueError("No dataframe available - cannot get columns")

        expected_data_type = (SparkDatatype.from_string(data_type) if isinstance(data_type, str) else data_type).value

        columns_of_given_type: List[str] = [
            col for col in self.df.columns if self.df.schema[col].dataType.typeName() == expected_data_type
        ]
        return columns_of_given_type

    def is_column_type_correct(self, column):
        """Check if column type is correct and handle it if not, when limit_data_type is set"""
        if not self.limit_data_type_is_set:
            return True

        if self.column_type_of_col(column) in (limit_data_types := self.get_limit_data_types()):
            return True

        # Raises a ValueError if the Column object is not of a given type and data_type_strict_mode is set
        if self.data_type_strict_mode_is_set:
            raise ValueError(
                f"Critical error: {column} is not of type {limit_data_types}. Exception is raised because "
                f"`data_type_strict_mode` is set to True for {self.name}."
            )

        # Otherwise, throws a warning that the Column object is not of a given type
        self.log.warning(f"Column `{column}` is not of type `{limit_data_types}` and will be skipped.")
        return False

    def get_limit_data_types(self):
        """Get the limit_data_type as a list of strings"""
        return [dt.value for dt in self.ColumnConfig.limit_data_type]

    def get_columns(self) -> iter:
        """Return an iterator of the columns"""
        # If `run_for_all_is_set` is True, we want to run the transformation for all columns of a given type
        if self.run_for_all_is_set:
            columns = []
            for data_type in self.ColumnConfig.run_for_all_data_type:
                columns += self.get_all_columns_of_specific_type(data_type)
        else:
            columns = self.columns

        for column in columns:
            if self.is_column_type_correct(column):
                yield column


class ColumnsTransformationWithTarget(ColumnsTransformation, ABC):
    """Extended ColumnsTransformation class with an additional `target_column` field

    Using this class makes implementing Transformations significantly easier.

    Concept
    -------
    A `ColumnsTransformationWithTarget` is a `ColumnsTransformation` with an additional `target_column` field. This
    field can be used to store the result of the transformation in a new column.

    If the `target_column` is not provided, the result will be stored in the source column.

    If more than one column is passed, the behavior of the Class changes this way:

    - the transformation will be run in a loop against all the given columns
    - automatically handles the renaming of the columns when more than one column is passed
    - the `target_column` will be used as a suffix. Leaving this blank will result in the original columns being renamed

    The `func` method should be implemented in the child class. This method should return the transformation that will
    be applied to the column(s). The execute method (already preset) will use the `get_columns_with_target` method to
    loop over all the columns and apply this function to transform the DataFrame.

    Parameters
    ----------
    columns : ListOfColumns, optional, default=*
        The column (or list of columns) to apply the transformation to. Alias: column. If not provided, the
        `run_for_all_data_type` will be used to determine the data type. If `run_for_all_data_type` is not set, the
        transformation will be run for all columns of a given type.
    target_column : Optional[str], optional, default=None
        The name of the column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this input will be used as a suffix instead.

    Example
    -------
    Writing your own transformation using the `ColumnsTransformationWithTarget` class:

    ```python
    from pyspark.sql import Column
    from koheesio.steps.transformations import ColumnsTransformationWithTarget


    class AddOneWithTarget(ColumnsTransformationWithTarget):
        def func(self, col: Column):
            return col + 1
    ```

    In the above example, the `func` method is implemented to add 1 to the values of a given column.

    In order to use this transformation, we can call the `transform` method:
    ```python
    from pyspark.sql import SparkSession

    # create a DataFrame with 3 rows
    df = SparkSession.builder.getOrCreate().range(3)

    output_df = AddOneWithTarget(column="id", target_column="new_id").transform(df)
    ```

    The `output_df` will now contain the original DataFrame with an additional column called `new_id` with the values of
    `id` + 1.

    __output_df:__

    |id|new_id|
    |--|------|
    | 0|     1|
    | 1|     2|
    | 2|     3|


    > Note: The `target_column` will be used as a suffix when more than one column is given as source. Leaving this
    blank will result in the original columns being renamed.
    """

    target_column: Optional[str] = Field(
        default=None,
        alias="target_suffix",
        description="The column to store the result in. If not provided, the result will be stored in the source"
        "column. Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix",
    )

    @abstractmethod
    def func(self, column: Column) -> Column:
        """The function that will be run on a single Column of the DataFrame

        The `func` method should be implemented in the child class. This method should return the transformation that
        will be applied to the column(s). The execute method (already preset) will use the `get_columns_with_target`
        method to loop over all the columns and apply this function to transform the DataFrame.

        Parameters
        ----------
        column : Column
            The column to apply the transformation to

        Returns
        -------
        Column
            The transformed column
        """
        raise NotImplementedError

    def get_columns_with_target(self) -> iter:
        """Return an iterator of the columns

        Works just like in get_columns from the  ColumnsTransformation class except that it handles the `target_column`
        as well.

        If more than one column is passed, the behavior of the Class changes this way:
        - the transformation will be run in a loop against all the given columns
        - the target_column will be used as a suffix. Leaving this blank will result in the original columns being
            renamed.

        Returns
        -------
        iter
            An iterator of tuples containing the target column name and the original column name
        """
        columns = [*self.get_columns()]

        for column in columns:
            # ensures that we at least use the original column name
            target_column = self.target_column or column

            if len(columns) > 1:  # target_column becomes a suffix when more than 1 column is given
                # dict.fromkeys is used to avoid duplicates in the name while maintaining order
                _cols = [column, target_column]
                target_column = "_".join(list(dict.fromkeys(_cols)))

            yield target_column, column

    def execute(self):
        """Execute on a ColumnsTransformationWithTarget handles self.df (input) and set self.output.df (output)
        This can be left unchanged, and hence should not be implemented in the child class.
        """
        df = self.df

        for target_column, column in self.get_columns_with_target():
            func = self.func  # select the applicable function
            df = df.withColumn(
                target_column,
                func(f.col(column)),
            )

        self.output.df = df



"""
Spark Utility functions
"""

import os
from enum import Enum

from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DataType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    NullType,
    ShortType,
    StringType,
    StructType,
    TimestampType,
)
from pyspark.version import __version__ as spark_version

__all__ = [
    "SparkDatatype",
    "get_spark_minor_version",
    "import_pandas_based_on_pyspark_version",
    "on_databricks",
    "schema_struct_to_schema_str",
    "spark_data_type_is_array",
    "spark_data_type_is_numeric",
    "spark_minor_version",
]


class SparkDatatype(Enum):
    """
    Allowed spark datatypes

    The following table lists the data types that are supported by Spark SQL.

    | Data type	    | SQL name                  |
    |---------------|---------------------------|
    | ByteType      | BYTE, TINYINT             |
    | ShortType     | SHORT, SMALLINT           |
    | IntegerType   | INT, INTEGER              |
    | LongType      | LONG, BIGINT              |
    | FloatType     | FLOAT, REAL               |
    | DoubleType    | DOUBLE                    |
    | DecimalType   | DECIMAL, DEC, NUMERIC     |
    | StringType    | STRING                    |
    | BinaryType    | BINARY                    |
    | BooleanType	| BOOLEAN                   |
    | TimestampType | TIMESTAMP, TIMESTAMP_LTZ  |
    | DateType      | DATE                      |
    | ArrayType     | ARRAY<element_type>       |
    | MapType       | MAP<key_type, value_type> |
    | NullType      | VOID                      |

    Not supported yet
    ----------------
    * __TimestampNTZType__
        TIMESTAMP_NTZ
    * __YearMonthIntervalType__
        INTERVAL YEAR, INTERVAL YEAR TO MONTH, INTERVAL MONTH
    * __DayTimeIntervalType__
        INTERVAL DAY, INTERVAL DAY TO HOUR, INTERVAL DAY TO MINUTE, INTERVAL DAY TO SECOND, INTERVAL HOUR,
        INTERVAL HOUR TO MINUTE, INTERVAL HOUR TO SECOND, INTERVAL MINUTE, INTERVAL MINUTE TO SECOND, INTERVAL SECOND

    See Also
    --------
    https://spark.apache.org/docs/latest/sql-ref-datatypes.html#supported-data-types
    """

    # byte
    BYTE = "byte"
    TINYINT = "byte"

    # short
    SHORT = "short"
    SMALLINT = "short"

    # integer
    INTEGER = "integer"
    INT = "integer"

    # long
    LONG = "long"
    BIGINT = "long"

    # float
    FLOAT = "float"
    REAL = "float"

    # timestamp
    TIMESTAMP = "timestamp"
    TIMESTAMP_LTZ = "timestamp"

    # decimal
    DECIMAL = "decimal"
    DEC = "decimal"
    NUMERIC = "decimal"

    DATE = "date"
    DOUBLE = "double"
    STRING = "string"
    BINARY = "binary"
    BOOLEAN = "boolean"
    ARRAY = "array"
    MAP = "map"
    VOID = "void"

    @property
    def spark_type(self) -> DataType:
        """Returns the spark type for the given enum value"""
        mapping_dict = {
            "byte": ByteType,
            "short": ShortType,
            "integer": IntegerType,
            "long": LongType,
            "float": FloatType,
            "double": DoubleType,
            "decimal": DecimalType,
            "string": StringType,
            "binary": BinaryType,
            "boolean": BooleanType,
            "timestamp": TimestampType,
            "date": DateType,
            "array": ArrayType,
            "map": MapType,
            "void": NullType,
        }
        return mapping_dict[self.value]

    @classmethod
    def from_string(cls, value: str) -> "SparkDatatype":
        """Allows for getting the right Enum value by simply passing a string value
        This method is not case-sensitive
        """
        return getattr(cls, value.upper())


def get_spark_minor_version() -> float:
    """Returns the minor version of the spark instance.

    For example, if the spark version is 3.3.2, this function would return 3.3
    """
    return float(".".join(spark_version.split(".")[:2]))


# short-hand for the get_spark_minor_version function
spark_minor_version: float = get_spark_minor_version()


def on_databricks() -> bool:
    """Retrieve if we're running on databricks or elsewhere"""
    dbr_version = os.getenv("DATABRICKS_RUNTIME_VERSION", None)
    return dbr_version is not None and dbr_version != ""


def spark_data_type_is_array(data_type: DataType) -> bool:
    """Check if the column's dataType is of type ArrayType"""
    return isinstance(data_type, ArrayType)


def spark_data_type_is_numeric(data_type: DataType) -> bool:
    """Check if the column's dataType is of type ArrayType"""
    return isinstance(data_type, (IntegerType, LongType, FloatType, DoubleType, DecimalType))


def schema_struct_to_schema_str(schema: StructType) -> str:
    """Converts a StructType to a schema str"""
    if not schema:
        return ""
    return ",\n".join([f"{field.name} {field.dataType.typeName().upper()}" for field in schema.fields])


def import_pandas_based_on_pyspark_version():
    """
    This function checks the installed version of PySpark and then tries to import the appropriate version of pandas.
    If the correct version of pandas is not installed, it raises an ImportError with a message indicating which version
    of pandas should be installed.
    """
    try:
        import pandas as pd

        pyspark_version = get_spark_minor_version()
        pandas_version = pd.__version__

        if (pyspark_version < 3.4 and pandas_version >= "2") or (pyspark_version >= 3.4 and pandas_version < "2"):
            raise ImportError(
                f"For PySpark {pyspark_version}, "
                f"please install Pandas version {'< 2' if pyspark_version < 3.4 else '>= 2'}"
            )

        return pd
    except ImportError as e:
        raise ImportError("Pandas module is not installed.") from e



"""
Test the Pad, LPad, RPad, PadBoth columns transformations
"""

import pytest

from koheesio.logger import LoggingFactory
from koheesio.models import ValidationError
from koheesio.spark.transformations.strings.pad import LPad, Pad, RPad

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)

data = [[1, 2, "hello"], [3, 4, "world"], [3, 4, None]]
schema = ["id", "id2", "c1"]
standard_input_values = dict(
    columns=["c1"],
    target_column="result",
    character="*",
    length=10,
)


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: Test Pad
            # input values
            standard_input_values,
            # expected data
            dict(
                Pad=["*****hello", "*****world", None],
                LPad=["*****hello", "*****world", None],
                RPad=["hello*****", "world*****", None],
            ),
        ),
        (
            # description: Test Pad - Less than original length
            # input values
            {**standard_input_values, "length": 3},
            # expected data
            dict(
                Pad=["hel", "wor", None],
                LPad=["hel", "wor", None],
                RPad=["hel", "wor", None],
            ),
        ),
        (
            # description: Test Pad - Less than original length
            # input values
            {**standard_input_values, "character": "*-"},
            # expected data
            dict(
                Pad=["*-*-*hello", "*-*-*world", None],
                LPad=["*-*-*hello", "*-*-*world", None],
                RPad=["hello*-*-*", "world*-*-*", None],
            ),
        ),
    ],
)
def test_happy_flow(input_values, expected, spark):
    input_df = spark.createDataFrame(data, schema)

    klz = Pad, LPad, RPad

    for kls in klz:
        log.info(f"Testing for {kls.__name__}")
        trim = kls(**input_values)
        output_df = trim.transform(input_df)
        target_column = trim.target_column

        # log equivalent of doing df.show()
        log.info(f"show output_df: \n{output_df._jdf.showString(20, 20, False)}")

        actual = [row[target_column] for row in output_df.select(target_column).collect()]
        assert actual == expected[kls.__name__]


def test_left_without_char():
    log.info("Testing unhappy flow for LPad")
    with pytest.raises(ValidationError) as e:
        _ = LPad(**{**standard_input_values, "character": ""})

    exception = e.value
    log.info(f"{exception = }")
    assert "at least 1 character" in str(exception)


def test_left_negative_length():
    log.info("Testing unhappy flow for LPad")
    with pytest.raises(ValidationError) as e:
        _ = LPad(**{**standard_input_values, "length": -5})

    exception = e.value
    log.info(f"{exception = }")
    assert "greater than 0" in str(exception)


# TODO: not supported yet
#
#     def test_get_result_nested(self):
#         df = (
#             Pad(
#                 df=self.sample_nested_df,
#                 params={
#                     "source_column": "payload.col2.array_col.item",
#                     "length": 10,
#                     "character": "0",
#                     "direction": "left",
#                 },
#                 target_alias="result_column",
#                 columns_to_flatten=["payload{}.col2{}.array_col[].item"],
#             )
#             .execute()
#             .df
#         )
#         df.printSchema()
#         self.assertEqual(df.collect()[0].result_column, "00000000v1")
#         self.assertEqual(df.collect()[1].result_column, "00000000v2")



from unittest import mock

import paramiko
import pytest
from paramiko import SSHException

from koheesio.integrations.spark.sftp import (
    SendCsvToSftp,
    SendJsonToSftp,
    SFTPWriteMode,
    SFTPWriter,
)
from koheesio.spark.writers.buffer import PandasCsvBufferWriter

pytestmark = pytest.mark.spark


@pytest.fixture
def buffer_writer(spark):
    # Create a small Spark DataFrame for testing
    df = spark.createDataFrame([(1, "a"), (2, "b")], ["column1", "column2"])
    return PandasCsvBufferWriter(df=df)


expected_data = "column1,column2\n1,a\n2,b\n"


@pytest.fixture(scope="session")
def sftp_fixture(sftpserver):
    # https://github.com/ulope/pytest-sftpserver/issues/30
    sftpserver.daemon_threads = True
    sftpserver.block_on_close = False
    yield sftpserver


@pytest.fixture(scope="function")
def prepare_sftp(sftp_fixture):
    # create empty files since the sftpserver doesn't allow to create and write to file
    with sftp_fixture.serve_content(
        {
            "test_dir": {
                "overwrite": {"test.csv": "existing_data"},
                "append": {"test.csv": "existing_data"},
                "ignore": {"test.csv": "existing_data"},
                "exclusive": {"test.csv": "existing_data"},
                "backup": {"test.csv": "existing_data"},
                "update_diff_data": {"test.csv": "existing_data"},
                "update_same_data": {"test.csv": expected_data},
            }
        }
    ):
        yield


def read_sftp_file(sftp_fixture, file_path):
    data = None
    sftp = None
    transport = None
    try:
        # Create a new transport and connect to the SFTP server
        transport = paramiko.Transport((sftp_fixture.host, sftp_fixture.port))
        transport.connect(username="a", password="b")

        # Create an SFTP client from the transport
        sftp = paramiko.SFTPClient.from_transport(transport)

        # Try to read the test data from the server
        with sftp.open(file_path, "r") as file:
            data = file.read().decode("utf-8")

    except Exception as e:
        print(f"An error occurred: {e}")

    finally:
        # Close the client and transport
        if sftp:
            sftp.close()
        if transport:
            transport.close()

    return data


def test_canary(sftp_fixture, prepare_sftp):
    """sftpserver setup canary test"""
    actual_data = read_sftp_file(sftp_fixture, "/test_dir/overwrite/test.csv")
    assert actual_data == "existing_data"


def test_invalid_write_mode():
    """Test that an error is raised for an invalid write mode"""
    with pytest.raises(ValueError):
        SFTPWriter(
            buffer_writer=PandasCsvBufferWriter(),
            mode="invalid_mode",
            host="localhost",
            port=22,
            username="a",
            password="b",
            path="/test.csv",
        )


def test_connection_failure(sftp_fixture, buffer_writer):
    """Test that an error is raised when the connection to the SFTP server fails"""
    with mock.patch("paramiko.Transport.connect", side_effect=SSHException("Could not connect")):
        with pytest.raises(SSHException):
            writer = SFTPWriter(
                buffer_writer=buffer_writer,
                host=sftp_fixture.host,
                port=sftp_fixture.port,
                username="a",
                password="b",
                path="/test.csv",
            )
            writer.execute()


def test_file_read_failure(sftp_fixture, buffer_writer):
    """Test that an error is raised when reading a file from the SFTP server fails"""
    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode="overwrite",
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path="/test.csv",
    )
    buffer_output = buffer_writer.output.read()
    with mock.patch.object(writer.client, "open", side_effect=IOError("Could not read file")):
        with pytest.raises(IOError):
            writer._handle_write_mode("/test.csv", buffer_output)


def test_file_write_failure(sftp_fixture, buffer_writer):
    """Test that an error is raised when writing a file to the SFTP server fails"""
    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path="/test.csv",
    )
    with mock.patch.object(writer, "write_file", side_effect=IOError("Could not write file")):
        with pytest.raises(IOError):
            writer.execute()


def test_execute_overwrite_mode(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/overwrite/test.csv"
    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.OVERWRITE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == expected_data


def test_execute_append_mode(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/append/test.csv"
    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.APPEND,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == "existing_data" + expected_data


def test_execute_ignore_mode_existing_file(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/ignore/test.csv"

    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.IGNORE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == "existing_data"  # The existing data should not be overwritten


def test_execute_exclusive_mode_existing_file(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/exclusive/test.csv"

    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.EXCLUSIVE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act & Assert
    with pytest.raises(FileExistsError):
        writer.execute()


def test_execute_exclusive_mode_new_file(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/exclusive/new_file.csv"

    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.EXCLUSIVE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == expected_data


def test_execute_backup_mode_existing_file(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/backup/test.csv"

    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.BACKUP,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == expected_data  # The existing data should be overwritten

    # Assert backup file is created
    transport = paramiko.Transport((sftp_fixture.host, sftp_fixture.port))
    transport.connect(username="a", password="b")

    sftp = None
    try:
        # Create an SFTP client from the transport
        sftp = paramiko.SFTPClient.from_transport(transport)

        # List the files in the directory
        files = sftp.listdir("/test_dir/backup")

        # Check if there is a file that matches the backup file pattern
        backup_files = [f for f in files if f.endswith(".bak")]
        assert len(backup_files) == 1

    finally:
        # Close the client and transport
        if sftp:
            sftp.close()
        transport.close()


def test_execute_update_mode_existing_file_same_data(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/update_same_data/test.csv"

    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.UPDATE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == expected_data  # The existing data should not be overwritten


def test_execute_update_mode_existing_file_different_data(sftp_fixture, prepare_sftp, buffer_writer):
    # SetUp
    path = "/test_dir/update_diff_data/test.csv"

    writer = SFTPWriter(
        buffer_writer=buffer_writer,
        mode=SFTPWriteMode.UPDATE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    # Act
    writer.execute()

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == expected_data  # The existing data should be overwritten


def test_write_method(sftp_fixture, prepare_sftp, spark):
    """Test that the write method works as expected"""
    # SetUp
    path = "/test_dir/overwrite/test.csv"

    writer = SFTPWriter(
        buffer_writer=PandasCsvBufferWriter(),  # We use a PandasCsvBufferWriter with no data and default parameters
        mode=SFTPWriteMode.OVERWRITE,
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
    )

    df = spark.createDataFrame([(1, "a"), (2, "b")], ["column1", "column2"])
    _expected_data = "column1,column2\n1,a\n2,b\n"

    # Act
    writer.write(df)

    # Assert
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == _expected_data


def test_write_mode_passed_as_string():
    """Test that the write mode can be passed as a string"""
    # SetUp
    writer = SFTPWriter(
        buffer_writer=PandasCsvBufferWriter(),
        mode="overwrite",
        host="localhost",
        port=22,
        username="a",
        password="b",
        path="/test.csv",
    )

    # Act & Assert
    assert writer.mode == SFTPWriteMode.OVERWRITE
    assert writer.write_mode == "wb"


def test_path_and_filename_process_correctly():
    """Test that the path and filename are processed correctly"""
    # SetUp
    writer = SFTPWriter(
        buffer_writer=PandasCsvBufferWriter(),
        mode="overwrite",
        host="goodbye",
        port=42,
        path="/test_dir/",
        filename="thanks_for_all_the_fish.csv",  # we use the alias for file_name explicitly
    )

    # Act & Assert
    assert writer.file_name == "thanks_for_all_the_fish.csv"
    assert writer.path.as_posix() == "/test_dir/thanks_for_all_the_fish.csv"


def test_sftp_removed_from_host_name():
    """Test that the SFTP server is removed from the host after the test"""
    # SetUp
    host = "sftp://localhost/"
    writer = SFTPWriter(
        buffer_writer=PandasCsvBufferWriter(),
        host=host,
        port=42,
        path="/test.csv",
    )

    # Assert
    assert writer.host == "localhost"


def test_invalid_path():
    """Test that an error is raised when an invalid path is passed"""
    with pytest.raises(ValueError):
        SFTPWriter(
            buffer_writer=PandasCsvBufferWriter(),
            host="localhost",
            port=22,
            path=123.456,
        )


def test_send_csv_to_sftp(sftp_fixture, prepare_sftp, spark):
    """Test that the write method works as expected"""
    # SetUp
    path = "/test_dir/overwrite/test.csv"
    test_data = [
        ("foo", 1),
        ("bar", 2),
        ("baz", 3),
    ]
    test_schema = ["name", "value"]

    writer = SendCsvToSftp(
        # Parameters for the SFTPWriter part
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
        # Parameters for the PandasCsvBufferWriter part
        sep="\t",
        line_terminator="\r",
        header=False,
        columns=test_schema,
    )

    _df = spark.createDataFrame(test_data, test_schema)
    _expected_data = "foo\t1\rbar\t2\rbaz\t3\r"

    # Act
    writer.write(_df)

    # Assert that the PandasCsvBufferWriter parameters are set correctly
    assert writer.buffer_writer.sep == "\t"
    assert writer.buffer_writer.lineSep == "\r"  # line_terminator is an alias for lineSep
    assert writer.buffer_writer.header is False
    assert writer.buffer_writer.columns == test_schema

    # Assert that the data was written to the SFTP server
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == _expected_data


def test_send_json_to_sftp(sftp_fixture, prepare_sftp, spark):
    """Test that the write method works as expected"""
    # SetUp
    path = "/test_dir/overwrite/test.csv"
    test_data = [
        ("foo", 1),
        ("bar", 2),
        ("baz", 3),
    ]
    test_schema = ["name", "value"]

    writer = SendJsonToSftp(
        # Parameters for the SFTPWriter part
        host=sftp_fixture.host,
        port=sftp_fixture.port,
        username="a",
        password="b",
        path=path,
        # Parameters for the PandasJsonBufferWriter part
        orient="split",
        columns=test_schema,
    )

    _df = spark.createDataFrame(test_data, test_schema)
    _expected_data = "{" '"columns":["name","value"],' '"index":[0,1,2],' '"data":[["foo",1],["bar",2],["baz",3]]' "}"

    # Act
    writer.write(_df)

    # Assert that the PandasCsvBufferWriter parameters are set correctly
    assert writer.buffer_writer.orient == "split"
    assert writer.buffer_writer.columns == test_schema
    assert writer.buffer_writer.compression is None

    # Assert that the data was written to the SFTP server
    actual_data = read_sftp_file(sftp_fixture, path)
    assert actual_data == _expected_data



"""
String transformations using regular expressions.

This module contains transformations that use regular expressions to transform strings.

Classes
-------
RegexpExtract
    Extract a specific group matched by a Java regexp from the specified string column.

RegexpReplace
    Searches for the given regexp and replaces all instances with what is in 'replacement'.

"""

from typing import Optional

from pyspark.sql import Column
from pyspark.sql.functions import regexp_extract, regexp_replace

from koheesio.models import Field
from koheesio.spark.transformations import ColumnsTransformationWithTarget


class RegexpExtract(ColumnsTransformationWithTarget):
    """
    Extract a specific group matched by a Java regexp from the specified string column.
    If the regexp did not match, or the specified group did not match, an empty string is returned.

    A wrapper around the pyspark regexp_extract function

    Parameters
    ----------
    columns : ListOfColumns
        The column (or list of columns) to extract from. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    regexp : str
        The Java regular expression to extract
    index : Optional[int], optional, default=0
        When there are more groups in the match, you can indicate which one you want.
        0 means the whole match. 1 and above are groups within that match.

    Example
    -------

    ### Extracting the year and week number from a string

    Let's say we have a column containing the year and week in a format like `Y## W#` and we would like to extract the
    week numbers.

    __input_df:__

    | YWK      |
    |----------|
    | 2020 W1  |
    | 2021 WK2 |

    ```python
    output_df = RegexpExtract(
        column="YWK",
        target_column="week_number",
        regexp="Y([0-9]+) ?WK?([0-9]+)",
        index=2,  # remember that this is 1-indexed! So 2 will get the week number in this example.
    ).transform(input_df)
    ```

    __output_df:__

    | YWK     | week_number |
    |---------|-------------|
    | 2020 W1 | 1           |
    | 2021 WK2| 2           |

    ### Using the same example, but extracting the year instead

    If you want to extract the year, you can use index=1.

    ```python
    output_df = RegexpExtract(
        column="YWK",
        target_column="year",
        regexp="Y([0-9]+) ?WK?([0-9]+)",
        index=1,  # remember that this is 1-indexed! So 1 will get the year in this example.
    ).transform(input_df)
    ```

    __output_df:__

    | YWK     | year |
    |---------|------|
    | 2020 W1 | 2020 |
    | 2021 WK2| 2021 |
    """

    regexp: str = Field(default=..., description="The Java regular expression to extract")
    index: Optional[int] = Field(
        default=0,
        description="When there are more groups in the match, you can indicate which one you want. "
        "0 means the whole match. 1 and above are groups within that match.",
    )

    def func(self, column: Column):
        return regexp_extract(column, self.regexp, self.index)


class RegexpReplace(ColumnsTransformationWithTarget):
    """
    Searches for the given regexp and replaces all instances with what is in 'replacement'.

    A wrapper around the pyspark regexp_replace function

    Parameters
    ----------
    columns : ListOfColumns
        The column (or list of columns) to replace in. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    regexp : str
        The regular expression to replace
    replacement : str
        String to replace matched pattern with.

    Examples
    --------
    __input_df:__
    | content    |
    |------------|
    | hello world|

    Let's say you want to replace 'hello'.

    ```python
    output_df = RegexpReplace(
        column="content",
        target_column="replaced",
        regexp="hello",
        replacement="gutentag",
    ).transform(input_df)
    ```

    __output_df:__
    | content    | replaced      |
    |------------|---------------|
    | hello world| gutentag world|

    """

    regexp: str = Field(default=..., description="The regular expression to replace")
    replacement: str = Field(
        default=...,
        description="String to replace matched pattern with.",
    )

    def func(self, column: Column):
        return regexp_replace(column, self.regexp, self.replacement)



import pytest

from pyspark.sql.dataframe import DataFrame

from koheesio.spark.readers.metastore import MetastoreReader

pytestmark = pytest.mark.spark


def test_metastore_reader(spark):
    df = MetastoreReader(table="klettern.delta_test_table").read()
    actual = df.head().asDict()
    expected = {"id": 0}

    assert isinstance(df, DataFrame)
    assert actual == expected



import pytest

from pyspark.sql import functions as F

from koheesio.spark.readers.dummy import DummyReader
from koheesio.spark.transformations.camel_to_snake import CamelToSnakeTransformation

pytestmark = pytest.mark.spark


class TestCamelToSnakeTransformation:
    @pytest.fixture
    def input_df(self):
        df = (
            DummyReader(range=1)
            .read()
            .withColumn("camelCaseColumn", F.col("id"))
            .withColumn("snake_case_column", F.col("id"))
        )
        yield df

    def test_transformation(self, input_df):
        transformation = CamelToSnakeTransformation()

        df = transformation.transform(input_df)
        actual = df.head().asDict()

        expected = {"id": 0, "camel_case_column": 0, "snake_case_column": 0}
        assert actual == expected

    def test_transform_on_specific_column_only(self, input_df):
        input_df = input_df.withColumn("anotherCamelCaseColumn", F.col("id"))
        transformation = CamelToSnakeTransformation(column="anotherCamelCaseColumn")

        df = transformation.transform(input_df)
        actual = df.head().asDict()

        expected = {"id": 0, "camelCaseColumn": 0, "snake_case_column": 0, "another_camel_case_column": 0}
        assert actual == expected



"""
Module for reading data from JDBC sources.

Classes
-------
JdbcReader
    Reader for JDBC tables.
"""

from typing import Any, Dict, Optional

from koheesio.models import Field, SecretStr
from koheesio.spark.readers import Reader


class JdbcReader(Reader):
    """
    Reader for JDBC tables.

    Wrapper around Spark's jdbc read format

    Notes
    -----
    * Query has precedence over dbtable. If query and dbtable both are filled in, dbtable will be ignored!
    * Extra options to the spark reader can be passed through the `options` input. Refer to Spark documentation
        for details: https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html
    * Consider using `fetchsize` as one of the options, as it is greatly increases the performance of the reader
    * Consider using `numPartitions`, `partitionColumn`, `lowerBound`, `upperBound` together with real or synthetic
        partitioning column as it will improve the reader performance

    When implementing a JDBC reader, the `get_options()` method should be implemented. The method should return a dict
    of options required for the specific JDBC driver. The `get_options()` method can be overridden in the child class.
    Additionally, the `driver` parameter should be set to the name of the JDBC driver. Be aware that the driver jar
    needs to be included in the Spark session; this class does not (and can not) take care of that!

    Example
    -------
    > Note: jars should be added to the Spark session manually. This class does not take care of that.

    This example depends on the jar for MS SQL:
     `https://repo1.maven.org/maven2/com/microsoft/sqlserver/mssql-jdbc/9.2.1.jre8/mssql-jdbc-9.2.1.jre8.jar`

    ```python
    from koheesio.spark.readers.jdbc import JdbcReader

    jdbc_mssql = JdbcReader(
        driver="com.microsoft.sqlserver.jdbc.SQLServerDriver",
        url="jdbc:sqlserver://10.xxx.xxx.xxx:1433;databaseName=YOUR_DATABASE",
        user="YOUR_USERNAME",
        password="***",
        dbtable="schemaname.tablename",
        options={"fetchsize": 100},
    )
    df = jdbc_mssql.read()
    ```
    """

    format: str = Field(default="jdbc", description="The type of format to load. Defaults to 'jdbc'.")
    driver: str = Field(
        default=...,
        description="Driver name. Be aware that the driver jar needs to be passed to the task",
    )
    url: str = Field(
        default=...,
        description="URL for the JDBC driver. Note, in some environments you need to use the IP Address instead of "
        "the hostname of the server.",
    )
    user: str = Field(default=..., description="User to authenticate to the server")
    password: SecretStr = Field(default=..., description="Password belonging to the username")
    dbtable: Optional[str] = Field(
        default=None, description="Database table name, also include schema name", alias="table"
    )
    query: Optional[str] = Field(default=None, description="Query")
    options: Optional[Dict[str, Any]] = Field(default_factory=dict, description="Extra options to pass to spark reader")

    def get_options(self):
        """
        Dictionary of options required for the specific JDBC driver.

        Note: override this method if driver requires custom names, e.g. Snowflake: `sfUrl`, `sfUser`, etc.
        """
        return {
            "driver": self.driver,
            "url": self.url,
            "user": self.user,
            "password": self.password,
            **self.options,
        }

    def execute(self):
        """Wrapper around Spark's jdbc read format"""

        # Can't have both dbtable and query empty
        if not self.dbtable and not self.query:
            raise ValueError("Please do not leave dbtable and query both empty!")

        if self.query and self.dbtable:
            self.log.info("Both 'query' and 'dbtable' are filled in, 'dbtable' will be ignored!")

        options = self.get_options()

        if pw := self.password:
            options["password"] = pw.get_secret_value()

        if query := self.query:
            options["query"] = query
            self.log.info(f"Executing query: {self.query}")
        else:
            options["dbtable"] = self.dbtable

        self.output.df = self.spark.read.format(self.format).options(**options).load()



from datetime import datetime
from unittest import mock

import chispa
import pytest
from conftest import await_job_completion

import pydantic

from pyspark.sql import DataFrame

from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.readers.delta import DeltaTableReader
from koheesio.spark.snowflake import (
    RunQuery,
    SnowflakeWriter,
    SynchronizeDeltaToSnowflakeTask,
)
from koheesio.spark.writers import BatchOutputMode, StreamingOutputMode
from koheesio.spark.writers.delta import DeltaTableWriter
from koheesio.spark.writers.stream import ForEachBatchStreamWriter

pytestmark = pytest.mark.spark

COMMON_OPTIONS = {
    "source_table": DeltaTableStep(table=""),
    "target_table": "foo.bar",
    "key_columns": [
        "Country",
    ],
    "url": "url",
    "user": "user",
    "password": "password",
    "database": "db",
    "schema": "schema",
    "role": "role",
    "warehouse": "warehouse",
    "persist_staging": False,
    "checkpoint_location": "some_checkpoint_location",
}


@pytest.fixture(scope="session")
def snowflake_staging_file(tmp_path_factory, random_uuid, logger):
    fldr = tmp_path_factory.mktemp("snowflake_staging.parq" + random_uuid)
    logger.debug(f"Building test checkpoint folder '{fldr}'")
    yield fldr.as_posix()


@pytest.fixture
def foreach_batch_stream_local(checkpoint_folder, snowflake_staging_file):
    def append_to_memory(df: DataFrame, batchId: int):
        df.write.mode("append").parquet(snowflake_staging_file)

    return ForEachBatchStreamWriter(
        output_mode=StreamingOutputMode.APPEND,
        batch_function=append_to_memory,
        checkpoint_location=checkpoint_folder,
    )


class TestSnowflakeSyncTask:
    @mock.patch.object(SynchronizeDeltaToSnowflakeTask, "writer")
    def test_overwrite(self, mock_writer, spark):
        source_table = DeltaTableStep(datbase="klettern", table="test_overwrite")

        df = spark.createDataFrame(
            data=[
                ("Australia", 100, 3000),
                ("USA", 10000, 20000),
                ("UK", 7000, 10000),
            ],
            schema=[
                "Country",
                "NumVaccinated",
                "AvailableDoses",
            ],
        )

        DeltaTableWriter(table=source_table, output_mode=BatchOutputMode.OVERWRITE, df=df).execute()

        task = SynchronizeDeltaToSnowflakeTask(
            streaming=False,
            synchronisation_mode=BatchOutputMode.OVERWRITE,
            **{**COMMON_OPTIONS, "source_table": source_table},
        )

        def mock_drop_table(table):
            pass

        with mock.patch.object(SynchronizeDeltaToSnowflakeTask, "drop_table") as mocked_drop_table:
            mocked_drop_table.return_value = mock_drop_table
            task.execute()
        # Ensure that this call doesn't raise an exception if called on a batch job
        task.writer.await_termination()
        chispa.assert_df_equality(task.output.target_df, df)

    @mock.patch.object(SynchronizeDeltaToSnowflakeTask, "writer")
    def test_overwrite_with_persist(self, mock_writer, spark):
        source_table = DeltaTableStep(datbase="klettern", table="test_overwrite")

        df = spark.createDataFrame(
            data=[
                ("Australia", 100, 3000),
                ("USA", 10000, 20000),
                ("UK", 7000, 10000),
            ],
            schema=[
                "Country",
                "NumVaccinated",
                "AvailableDoses",
            ],
        )

        DeltaTableWriter(table=source_table, output_mode=BatchOutputMode.OVERWRITE, df=df).execute()

        task = SynchronizeDeltaToSnowflakeTask(
            streaming=False,
            synchronisation_mode=BatchOutputMode.OVERWRITE,
            **{**COMMON_OPTIONS, "source_table": source_table, "persist_staging": True},
        )

        def mock_drop_table(table):
            pass

        task.execute()
        chispa.assert_df_equality(task.output.target_df, df)

    @mock.patch.object(RunQuery, "execute")
    def test_merge(
        self,
        mocked_sf_query_execute,
        spark,
        foreach_batch_stream_local,
        snowflake_staging_file,
    ):
        # Prepare Delta requirements
        source_table = DeltaTableStep(datbase="klettern", table="test_merge")
        spark.sql(
            f"""
        CREATE OR REPLACE TABLE {source_table.table_name}
        (Country STRING, NumVaccinated LONG, AvailableDoses LONG)
        USING DELTA
        TBLPROPERTIES ('delta.enableChangeDataFeed' = true);
        """
        )

        # Prepare local representation of snowflake
        task = SynchronizeDeltaToSnowflakeTask(
            streaming=True,
            synchronisation_mode=BatchOutputMode.MERGE,
            **{**COMMON_OPTIONS, "source_table": source_table},
        )

        # Perform actions
        spark.sql(
            f"""INSERT INTO {source_table.table_name} VALUES
            ("Australia", 100, 3000),
            ("USA", 10000, 20000),
            ("UK", 7000, 10000);
        """
        )

        # Run code

        with mock.patch.object(SynchronizeDeltaToSnowflakeTask, "writer", new=foreach_batch_stream_local):
            task.execute()
            task.writer.await_termination()

        # Validate result
        df = spark.read.parquet(snowflake_staging_file).select("Country", "NumVaccinated", "AvailableDoses")

        chispa.assert_df_equality(
            df,
            spark.sql(f"SELECT * FROM {source_table.table_name}"),
            ignore_row_order=True,
            ignore_column_order=True,
        )
        assert df.count() == 3

        # Perform update
        spark.sql(f"""INSERT INTO {source_table.table_name} VALUES ("BELGIUM", 10, 100)""")
        spark.sql(f"UPDATE {source_table.table_name} SET NumVaccinated = 20 WHERE Country = 'Belgium'")

        # Run code
        with mock.patch.object(SynchronizeDeltaToSnowflakeTask, "writer", new=foreach_batch_stream_local):
            # Test that this call doesn't raise exception after all queries were completed
            task.writer.await_termination()
            task.execute()
            await_job_completion()

        # Validate result
        df = spark.read.parquet(snowflake_staging_file).select("Country", "NumVaccinated", "AvailableDoses")

        chispa.assert_df_equality(
            df,
            spark.sql(f"SELECT * FROM {source_table.table_name}"),
            ignore_row_order=True,
            ignore_column_order=True,
        )
        assert df.count() == 4

    def test_writer(self, spark):
        source_table = DeltaTableStep(datbase="klettern", table="test_overwrite")
        df = spark.createDataFrame(
            data=[
                ("Australia", 100, 3000),
                ("USA", 10000, 20000),
                ("UK", 7000, 10000),
            ],
            schema=[
                "Country",
                "NumVaccinated",
                "AvailableDoses",
            ],
        )

        DeltaTableWriter(table=source_table, output_mode=BatchOutputMode.OVERWRITE, df=df).execute()

        task = SynchronizeDeltaToSnowflakeTask(
            streaming=False,
            synchronisation_mode=BatchOutputMode.OVERWRITE,
            **{**COMMON_OPTIONS, "source_table": source_table},
        )

        assert task.writer is task.writer

    @pytest.mark.parametrize(
        "output_mode,streaming",
        [(BatchOutputMode.MERGE, True), (BatchOutputMode.APPEND, True), (BatchOutputMode.OVERWRITE, False)],
    )
    def test_schema_tracking_location(self, output_mode, streaming):
        source_table = DeltaTableStep(datbase="klettern", table="test_overwrite")

        task = SynchronizeDeltaToSnowflakeTask(
            streaming=streaming,
            synchronisation_mode=output_mode,
            schema_tracking_location="/schema/tracking/location",
            **{**COMMON_OPTIONS, "source_table": source_table},
        )

        reader = task.reader
        assert reader.schema_tracking_location == "/schema/tracking/location"


class TestMerge:
    def test_non_key_columns(self, spark):
        table = DeltaTableStep(database="klettern", table="sync_test_table")
        spark.sql(
            f"""
        CREATE OR REPLACE TABLE {table.table_name}
        (Country STRING, NumVaccinated INT, AvailableDoses INT)
        USING DELTA
        TBLPROPERTIES ('delta.enableChangeDataFeed' = true);
        """
        )

        df = spark.createDataFrame(
            data=[
                (
                    "Australia",
                    100,
                    3000,
                    "insert",
                    2,
                    datetime(2021, 4, 14, 20, 26, 37),
                ),
                (
                    "USA",
                    10000,
                    20000,
                    "update_preimage",
                    3,
                    datetime(2021, 4, 14, 20, 26, 39),
                ),
                (
                    "USA",
                    11000,
                    20000,
                    "update_postimage",
                    3,
                    datetime(2021, 4, 14, 20, 26, 39),
                ),
                ("UK", 7000, 10000, "delete", 4, datetime(2021, 4, 14, 20, 26, 40)),
            ],
            schema=[
                "Country",
                "NumVaccinated",
                "AvailableDoses",
                "_change_type",
                "_commit_version",
                "_commit_timestamp",
            ],
        )
        with mock.patch.object(DeltaTableReader, "read") as mocked_read:
            mocked_read.return_value = df
            task = SynchronizeDeltaToSnowflakeTask(
                streaming=False,
                synchronisation_mode=BatchOutputMode.APPEND,
                **{**COMMON_OPTIONS, "source_table": table},
            )
            assert task.non_key_columns == ["NumVaccinated", "AvailableDoses"]

    def test_changed_table(self, spark, sample_df_with_timestamp):
        # Example CDF dataframe from https://docs.databricks.com/en/_extras/notebooks/source/delta/cdf-demo.html
        df = spark.createDataFrame(
            data=[
                (
                    "Australia",
                    100,
                    3000,
                    "insert",
                    2,
                    datetime(2021, 4, 14, 20, 26, 37),
                ),
                (
                    "USA",
                    10000,
                    20000,
                    "update_preimage",
                    3,
                    datetime(2021, 4, 14, 20, 26, 39),
                ),
                (
                    "USA",
                    11000,
                    20000,
                    "update_postimage",
                    3,
                    datetime(2021, 4, 14, 20, 26, 39),
                ),
                ("UK", 7000, 10000, "delete", 4, datetime(2021, 4, 14, 20, 26, 40)),
            ],
            schema=[
                "Country",
                "NumVaccinated",
                "AvailableDoses",
                "_change_type",
                "_commit_version",
                "_commit_timestamp",
            ],
        )

        expected_staging_df = spark.createDataFrame(
            data=[
                ("Australia", 100, 3000, "insert"),
                ("USA", 11000, 20000, "update_postimage"),
                ("UK", 7000, 10000, "delete"),
            ],
            schema=[
                "Country",
                "NumVaccinated",
                "AvailableDoses",
                "_change_type",
            ],
        )

        result_df = SynchronizeDeltaToSnowflakeTask._compute_latest_changes_per_pk(
            df, ["Country"], ["NumVaccinated", "AvailableDoses"]
        )

        chispa.assert_df_equality(
            result_df,
            expected_staging_df,
            ignore_row_order=True,
            ignore_column_order=True,
        )


class TestValidations:
    @pytest.mark.parametrize(
        "sync_mode,streaming",
        [
            (BatchOutputMode.OVERWRITE, False),
            (BatchOutputMode.MERGE, True),
            (BatchOutputMode.APPEND, False),
            (BatchOutputMode.APPEND, True),
        ],
    )
    def test_snowflake_sync_task_allowed_options(self, sync_mode: BatchOutputMode, streaming: bool):
        task = SynchronizeDeltaToSnowflakeTask(
            streaming=streaming,
            synchronisation_mode=sync_mode,
            **COMMON_OPTIONS,
        )

        assert task.reader.streaming == streaming

    @pytest.mark.parametrize(
        "sync_mode,streaming",
        [
            (BatchOutputMode.OVERWRITE, True),
            (BatchOutputMode.MERGE, False),
        ],
    )
    def test_snowflake_sync_task_unallowed_options(self, sync_mode: BatchOutputMode, streaming: bool):
        with pytest.raises(pydantic.ValidationError):
            SynchronizeDeltaToSnowflakeTask(
                streaming=streaming,
                synchronisation_mode=sync_mode,
                **COMMON_OPTIONS,
            )

    def test_snowflake_sync_task_merge_keys(self):
        with pytest.raises(pydantic.ValidationError):
            SynchronizeDeltaToSnowflakeTask(
                streaming=True,
                synchronisation_mode=BatchOutputMode.MERGE,
                **{**COMMON_OPTIONS, "key_columns": []},
            )

    @pytest.mark.parametrize(
        "sync_mode, streaming, expected_writer_type",
        [
            (BatchOutputMode.OVERWRITE, False, SnowflakeWriter),
            (BatchOutputMode.MERGE, True, ForEachBatchStreamWriter),
            (BatchOutputMode.APPEND, False, SnowflakeWriter),
            (BatchOutputMode.APPEND, True, ForEachBatchStreamWriter),
        ],
    )
    def test_snowflake_sync_task_allowed_writers(
        self, sync_mode: BatchOutputMode, streaming: bool, expected_writer_type: type
    ):
        # Overload dynamic retrieval of source schema
        with mock.patch.object(
            SynchronizeDeltaToSnowflakeTask,
            "non_key_columns",
            new=["NumVaccinated", "AvailableDoses"],
        ):
            task = SynchronizeDeltaToSnowflakeTask(
                streaming=streaming,
                synchronisation_mode=sync_mode,
                **COMMON_OPTIONS,
            )
            print(f"{task.writer = }")
            print(f"{type(task.writer) = }")
            assert isinstance(task.writer, expected_writer_type)

    def test_merge_cdf_enabled(self, spark):
        table = DeltaTableStep(database="klettern", table="sync_test_table")
        spark.sql(
            f"""
        CREATE OR REPLACE TABLE {table.table_name}
        (Country STRING, NumVaccinated INT, AvailableDoses INT)
        USING DELTA
        TBLPROPERTIES ('delta.enableChangeDataFeed' = false);
        """
        )
        task = SynchronizeDeltaToSnowflakeTask(
            streaming=True,
            synchronisation_mode=BatchOutputMode.MERGE,
            **{**COMMON_OPTIONS, "source_table": table},
        )
        assert task.source_table.is_cdf_active is False

        # Fail if ChangeDataFeed is not enabled
        with pytest.raises(RuntimeError):
            task.execute()


class TestMergeQuery:
    def test_merge_query_no_delete(self):
        query = SynchronizeDeltaToSnowflakeTask._build_sf_merge_query(
            target_table="target_table",
            stage_table="tmp_table",
            pk_columns=["Country"],
            non_pk_columns=["NumVaccinated", "AvailableDoses"],
        )
        expected_query = """
        MERGE INTO target_table target
        USING tmp_table temp ON target.Country = temp.Country
        WHEN MATCHED AND temp._change_type = 'update_postimage' THEN UPDATE SET NumVaccinated = temp.NumVaccinated, AvailableDoses = temp.AvailableDoses
        WHEN NOT MATCHED AND temp._change_type != 'delete' THEN INSERT (Country, NumVaccinated, AvailableDoses) VALUES (temp.Country, temp.NumVaccinated, temp.AvailableDoses)
        """

        assert query == expected_query

    def test_merge_query_with_delete(self):
        query = SynchronizeDeltaToSnowflakeTask._build_sf_merge_query(
            target_table="target_table",
            stage_table="tmp_table",
            pk_columns=["Country"],
            non_pk_columns=["NumVaccinated", "AvailableDoses"],
            enable_deletion=True,
        )
        expected_query = """
        MERGE INTO target_table target
        USING tmp_table temp ON target.Country = temp.Country
        WHEN MATCHED AND temp._change_type = 'update_postimage' THEN UPDATE SET NumVaccinated = temp.NumVaccinated, AvailableDoses = temp.AvailableDoses
        WHEN NOT MATCHED AND temp._change_type != 'delete' THEN INSERT (Country, NumVaccinated, AvailableDoses) VALUES (temp.Country, temp.NumVaccinated, temp.AvailableDoses)
        WHEN MATCHED AND temp._change_type = 'delete' THEN DELETE"""

        assert query == expected_query

    def test_default_staging_table(self):
        task = SynchronizeDeltaToSnowflakeTask(
            streaming=True,
            synchronisation_mode=BatchOutputMode.MERGE,
            **{
                **COMMON_OPTIONS,
                "source_table": DeltaTableStep(database="klettern", table="sync_test_table"),
            },
        )

        assert task.staging_table == "sync_test_table_stg"

    def test_custom_staging_table(self):
        task = SynchronizeDeltaToSnowflakeTask(
            streaming=True,
            synchronisation_mode=BatchOutputMode.MERGE,
            staging_table_name="staging_table",
            **{
                **COMMON_OPTIONS,
                "source_table": DeltaTableStep(database="klettern", table="sync_test_table"),
            },
        )

        assert task.staging_table == "staging_table"

    def test_invalid_staging_table(self):
        with pytest.raises(ValueError):
            SynchronizeDeltaToSnowflakeTask(
                streaming=True,
                synchronisation_mode=BatchOutputMode.MERGE,
                staging_table_name="import.staging_table",
                **{
                    **COMMON_OPTIONS,
                    "source_table": DeltaTableStep(database="klettern", table="sync_test_table"),
                },
            )



import pytest
import requests
from requests import HTTPError
from requests.adapters import HTTPAdapter
from requests.exceptions import RetryError
from requests_mock.mocker import Mocker
from urllib3 import Retry

from koheesio.steps.http import (
    HttpDeleteStep,
    HttpGetStep,
    HttpMethod,
    HttpPostStep,
    HttpPutStep,
    HttpStep,
)

BASE_URL = "https://httpbin.org"
GET_ENDPOINT = f"{BASE_URL}/get"
POST_ENDPOINT = f"{BASE_URL}/post"
PUT_ENDPOINT = f"{BASE_URL}/put"
DELETE_ENDPOINT = f"{BASE_URL}/delete"
STATUS_404_ENDPOINT = f"{BASE_URL}/status/404"
STATUS_500_ENDPOINT = f"{BASE_URL}/status/500"
STATUS_503_ENDPOINT = f"{BASE_URL}/status/503"


@pytest.mark.parametrize(
    "endpoint,step,method,return_value,expected_status_code",
    [
        pytest.param(
            GET_ENDPOINT,
            HttpGetStep,
            "get",
            {"job_id": "1234"},
            200,
            id="httpGetStep_success",
        ),
        pytest.param(GET_ENDPOINT, HttpGetStep, "get", {"message": "Error Code"}, 400, id="httpGetStep_error"),
        pytest.param(GET_ENDPOINT, HttpGetStep, "get", None, 404, id="httpGetStep_emptyresponse"),
        pytest.param(
            POST_ENDPOINT,
            HttpPostStep,
            "post",
            {"job_id": "1234"},
            200,
            id="httpPostStep_success",
        ),
        pytest.param(POST_ENDPOINT, HttpPostStep, "post", {"message": "Error Code"}, 400, id="httpPosStep_error"),
        pytest.param(
            PUT_ENDPOINT,
            HttpPutStep,
            "put",
            {"detail": "Updated"},
            202,
            id="httpPutStep_success",
        ),
        pytest.param(PUT_ENDPOINT, HttpPutStep, "put", {"message": "Error Code"}, 400, id="httpPutStep_error"),
        pytest.param(
            PUT_ENDPOINT, HttpPutStep, "put", {"message": "Validation Error"}, 422, id="httpPutStep_validationError"
        ),
        pytest.param(
            DELETE_ENDPOINT,
            HttpDeleteStep,
            "delete",
            {"detail": "Deleted"},
            204,
            id="httpDeleteStep_success",
        ),
        pytest.param(
            DELETE_ENDPOINT, HttpDeleteStep, "delete", {"message": "Error Code"}, 400, id="httpDeleteStep_error"
        ),
        pytest.param(
            DELETE_ENDPOINT,
            HttpDeleteStep,
            "delete",
            {"message": "Validation Error"},
            422,
            id="httpDeleteStep_validationError",
        ),
    ],
)
def test_http_step(endpoint, step, method, return_value, expected_status_code):
    """
    Unit Testing the GET functions.
    Above parameters are for the success and failed GET API calls
    """

    with Mocker() as rm:
        rm.request(method=method, url=endpoint, json=return_value, status_code=int(expected_status_code))
        step = step(
            url=endpoint,
            headers={
                "Authorization": "Bearer token",
                "Content-Type": "application/json",
            },
        )
        if expected_status_code not in (200, 202, 204):
            with pytest.raises(HTTPError) as excinfo:
                step.execute()

            assert excinfo.value.response.status_code == expected_status_code
        else:
            step.execute()
            assert step.output.status_code == expected_status_code  # type: ignore
            assert step.output.response_json == return_value  # type: ignore


def test_http_step_with_valid_http_method():
    """
    Unit Testing the GET functions.
    Above parameters are for the success and failed GET API calls
    """

    with Mocker() as rm:
        rm.get(GET_ENDPOINT, status_code=int(200))
        response = HttpStep(method="get", url=GET_ENDPOINT).execute()
        assert response.status_code == 200


def test_http_step_with_invalid_http_method():
    with Mocker() as rm:
        rm.get(GET_ENDPOINT, status_code=int(200))
        # Will be raised during class instantiation
        with pytest.raises(AttributeError):
            HttpStep(method="foo", url=GET_ENDPOINT).execute()


def test_http_step_request():
    with Mocker() as rm:
        rm.put(PUT_ENDPOINT, status_code=int(200))
        # The default method for HttpStep class is GET, however the method specified in `request` options is PUT and
        # it will override the default.
        response = HttpStep(url=PUT_ENDPOINT).request(method=HttpMethod.PUT)
        assert response.status_code == 200


def test_get_headers():
    # Authorization headers are being converted into SecretStr under the hood to avoid dumping any
    # sensitive content into logs. However, when calling the `get_headers` method, the SecretStr is being
    # converted back to string, otherwise sensitive info would have looked like '**********'.
    actual_headers = HttpStep(
        url=GET_ENDPOINT,
        headers={"Authorization": "Bearer token", "Content-Type": "application/json"},
    ).get_headers()

    assert actual_headers["Authorization"] != "**********"
    assert actual_headers["Authorization"] == "Bearer token"
    assert actual_headers["Content-Type"] == "application/json"


@pytest.mark.parametrize(
    "max_retries, endpoint, status_code, expected_count, error_type",
    [
        pytest.param(4, STATUS_503_ENDPOINT, 503, 5, RetryError, id="max_retries_4_503"),
        pytest.param(7, STATUS_404_ENDPOINT, 404, 8, RetryError, id="max_retries_7_404"),
        pytest.param(17, STATUS_500_ENDPOINT, 500, 18, RetryError, id="max_retries_17_500"),
        pytest.param(17, STATUS_404_ENDPOINT, 503, 1, HTTPError, id="max_retries_17_404"),
    ],
)
def test_max_retries(max_retries, endpoint, status_code, expected_count, error_type):
    session = requests.Session()
    retry_logic = Retry(total=max_retries, status_forcelist=[status_code])
    session.mount("https://", HTTPAdapter(max_retries=retry_logic))
    session.mount("http://", HTTPAdapter(max_retries=retry_logic))

    step = HttpGetStep(url=endpoint, session=session)

    with pytest.raises(error_type):
        step.execute()

    first_pool = [v for _, v in session.adapters["https://"].poolmanager.pools._container.items()][0]

    assert first_pool.num_requests == expected_count


@pytest.mark.parametrize(
    "backoff,expected",
    [
        pytest.param(7, [0, 14, 28], id="backoff_7"),
        pytest.param(3, [0, 6, 12], id="backoff_3"),
        pytest.param(2, [0, 4, 8], id="backoff_2"),
        pytest.param(1, [0, 2, 4], id="backoff_1"),
    ],
)
def test_initial_delay_and_backoff(monkeypatch, backoff, expected):
    session = requests.Session()
    retry_logic = Retry(total=3, backoff_factor=backoff, status_forcelist=[503])
    session.mount("https://", HTTPAdapter(max_retries=retry_logic))
    session.mount("http://", HTTPAdapter(max_retries=retry_logic))

    step = HttpGetStep(
        url=STATUS_503_ENDPOINT,
        headers={
            "Authorization": "Bearer token",
            "Content-Type": "application/json",
        },
        session=session,
    )

    # Save the original get_backoff_time method
    original_get_backoff_time = Retry.get_backoff_time

    backoff_values = []

    # Define a new function to replace get_backoff_time
    def mock_get_backoff_time(self, *args, **kwargs):
        result = original_get_backoff_time(self, *args, **kwargs)
        backoff_values.append(result)
        return result

    # Use monkeypatch to replace get_backoff_time
    monkeypatch.setattr(Retry, "get_backoff_time", mock_get_backoff_time)

    with pytest.raises(RetryError):
        step.execute()

    # Restore the original get_backoff_time method
    monkeypatch.undo()

    assert backoff_values == expected



import datetime
import os
import sys
from decimal import Decimal
from pathlib import Path
from textwrap import dedent
from unittest.mock import Mock

import pytest
from delta import configure_spark_with_delta_pip

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    NullType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from koheesio.logger import LoggingFactory
from koheesio.spark.readers.dummy import DummyReader


@pytest.fixture(scope="session")
def warehouse_path(tmp_path_factory, random_uuid, logger):
    fldr = tmp_path_factory.mktemp("spark-warehouse" + random_uuid)
    logger.debug(f"Building test warehouse folder '{fldr}'")
    yield fldr.as_posix()


@pytest.fixture(scope="function")
def checkpoint_folder(tmp_path_factory, random_uuid, logger):
    fldr = tmp_path_factory.mktemp("checkpoint" + random_uuid)
    logger.debug(f"Building test checkpoint folder '{fldr}'")
    yield fldr.as_posix()


@pytest.fixture(scope="session")
def spark(warehouse_path, random_uuid):
    """Spark session fixture with Delta enabled."""
    builder = (
        SparkSession.builder.appName("test_session" + random_uuid)
        .master("local[*]")
        .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config("spark.sql.warehouse.dir", warehouse_path)
        .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
        .config("spark.sql.session.timeZone", "UTC")
        .config("spark.sql.execution.arrow.pyspark.enabled", "true")
        .config("spark.sql.execution.arrow.pyspark.fallback.enabled", "true")
    )

    spark_session = configure_spark_with_delta_pip(builder).getOrCreate()
    yield spark_session

    spark_session.stop()


@pytest.fixture(autouse=True, scope="session")
def set_env_vars():
    """
    Set environment variables for PYSPARK_PYTHON and PYSPARK_DRIVER_PYTHON.

    This function sets the environment variables PYSPARK_PYTHON and PYSPARK_DRIVER_PYTHON
    to the current Python executable. It is used to ensure that the correct Python version
    is used when running PySpark.
    """
    existing_pyspark_python = os.environ.get("PYSPARK_PYTHON")
    existing_pyspark_driver_python = os.environ.get("PYSPARK_DRIVER_PYTHON")

    os.environ["PYSPARK_PYTHON"] = sys.executable
    os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable

    yield

    if existing_pyspark_python is not None:
        os.environ["PYSPARK_PYTHON"] = existing_pyspark_python
    else:
        del os.environ["PYSPARK_PYTHON"]

    if existing_pyspark_driver_python is not None:
        os.environ["PYSPARK_DRIVER_PYTHON"] = existing_pyspark_driver_python
    else:
        del os.environ["PYSPARK_DRIVER_PYTHON"]


@pytest.fixture(scope="session", autouse=True)
def setup(spark, delta_file):
    db_name = "klettern"

    if not spark.catalog.databaseExists(db_name):
        spark.sql(f"CREATE DATABASE {db_name}")
        spark.sql(f"USE {db_name}")

    setup_test_data(spark=spark, delta_file=Path(delta_file))
    yield


@pytest.fixture(scope="class")
def dummy_df(spark):
    """
    | id |
    |----|
    | 1  |
    """
    yield DummyReader(range=1).read()


@pytest.fixture(scope="class")
def sample_df_to_partition(spark):
    """
    | paritition | Value
    |----|----|
    | BE | 12 |
    | FR | 20 |
    """
    data = [["BE", 12], ["FR", 20]]
    schema = ["partition", "value"]
    return spark.createDataFrame(data, schema)


@pytest.fixture
def streaming_dummy_df(spark, delta_file):
    setup_test_data(spark=spark, delta_file=Path(delta_file))
    yield spark.readStream.table("delta_test_table")


@pytest.fixture(scope="class")
def sample_df_with_strings(spark):
    """
    df:
    | id | string |
    |----|--------|
    | 1  | hello  |
    | 2  | world  |
    | 3  |        |
    """
    data = [[1, "hello"], [2, "world"], [3, ""]]
    schema = ["id", "string"]
    return spark.createDataFrame(data, schema)


@pytest.fixture(scope="class")
def sample_df_with_timestamp(spark):
    """
    df:
    | id | a_date              | a_timestamp
    |----|---------------------|---------------------
    | 1  | 1970-04-20 12:33:09 |
    | 2  | 1980-05-21 13:34:08 |
    | 3  | 1990-06-22 14:35:07 |

    Schema:
    - id: bigint (nullable = true)
    - date: timestamp (nullable = true)
    """
    data = [
        (1, datetime.datetime(1970, 4, 20, 12, 33, 9), datetime.datetime(2000, 7, 1, 1, 1)),
        (2, datetime.datetime(1980, 5, 21, 13, 34, 8), datetime.datetime(2010, 8, 2, 2, 2)),
        (3, datetime.datetime(1990, 6, 22, 14, 35, 7), datetime.datetime(2020, 9, 3, 3, 3)),
    ]
    schema = ["id", "a_date", "a_timestamp"]
    return spark.createDataFrame(data, schema)


@pytest.fixture(scope="class")
def sample_df_with_string_timestamp(spark):
    """
    | id | a_string_timestamp |
    |----|--------------------|
    | 1  | 202301010420       |
    | 2  | 202302020314       |
    """
    data = [(1, "202301010420"), (2, "202302020314")]
    schema = ["id", "a_string_timestamp"]
    return spark.createDataFrame(data, schema)


def setup_test_data(spark, delta_file):
    """
    Sets up test data for the Spark session. Reads a Delta file, creates a temporary view,
    and populates a Delta table with the view's data.
    """
    delta_file = delta_file.absolute().as_posix()
    spark.read.format("delta").load(delta_file).limit(10).createOrReplaceTempView("delta_test_view")
    spark.sql(
        dedent(
            """
            CREATE TABLE IF NOT EXISTS delta_test_table
            USING DELTA
            TBLPROPERTIES ("delta.enableChangeDataFeed" = "true")
            AS SELECT v.* FROM delta_test_view v
            """
        )
    )


@pytest.fixture(scope="class")
def dummy_spark():
    class DummySpark:
        """Mocking SparkSession"""

        def __init__(self):
            self.options_dict = {}

        def mock_method(self, *args, **kwargs):
            return self

        @property
        def mock_property(self):
            return self

        def mock_options(self, *args, **kwargs):
            self.options_dict = kwargs
            return self

        options = mock_options
        format = mock_method
        read = mock_property

        _jvm = Mock()
        _jvm.net.snowflake.spark.snowflake.Utils.runQuery.return_value = True

        @staticmethod
        def load() -> DataFrame:
            df = Mock(spec=DataFrame)
            df.count.return_value = 1
            df.schema = StructType([StructField("foo", StringType(), True)])
            return df

    return DummySpark()


def await_job_completion(timeout=300, query_id=None):
    """
    Waits for a Spark streaming job to complete.

    This function checks the status of a Spark streaming job and waits until it is completed or a timeout is reached.
    If a query_id is provided, it waits for the specific streaming job with that id. Otherwise, it waits for any active
    streaming job.
    """
    logger = LoggingFactory.get_logger(name="await_job_completion", inherit_from_koheesio=True)

    start_time = datetime.datetime.now()
    spark = SparkSession.getActiveSession()
    logger.info("Waiting for streaming job to complete")
    if query_id is not None:
        stream = spark.streams.get(query_id)
        while stream.isActive and (datetime.datetime.now() - start_time).seconds < timeout:
            spark.streams.awaitAnyTermination(20)
    else:
        while len(spark.streams.active) > 0 and (datetime.datetime.now() - start_time).seconds < timeout:
            spark.streams.awaitAnyTermination(20)
    spark.streams.resetTerminated()
    logger.info("Streaming job completed")


@pytest.fixture
def df_with_all_types(spark):
    """Create a DataFrame with all supported Spark datatypes

    This DataFrame has 1 row and 14 columns, each column containing a single value of the given datatype.
    """
    data = dict(
        BYTE=(1, "byte", ByteType()),
        SHORT=(1, "short", ShortType()),
        INTEGER=(1, "integer", IntegerType()),
        LONG=(1, "long", LongType()),
        FLOAT=(1.0, "float", FloatType()),
        DOUBLE=(1.0, "double", DoubleType()),
        DECIMAL=(Decimal(1.0), "decimal", DecimalType()),
        STRING=("a", "string", StringType()),
        BINARY=(b"a", "binary", BinaryType()),
        BOOLEAN=(True, "boolean", BooleanType()),
        # '2023-01-01T00:01:01'
        TIMESTAMP=(datetime.datetime.utcfromtimestamp(1672531261), "timestamp", TimestampType()),
        DATE=(datetime.date(2023, 1, 1), "date", DateType()),
        ARRAY=(["a"], "array", ArrayType(StringType())),
        MAP=({"a": "b"}, "map", MapType(StringType(), StringType())),
        VOID=(None, "void", NullType()),
    )
    return spark.createDataFrame(
        data=[[v[0] for v in data.values()]],
        schema=StructType([StructField(name=v[1], dataType=v[2]) for v in data.values()]),
    )



"""
This module defines the DropColumn class, a subclass of ColumnsTransformation.
"""

from koheesio.spark.transformations import ColumnsTransformation


class DropColumn(ColumnsTransformation):
    """Drop one or more columns

    The DropColumn class is used to drop one or more columns from a PySpark DataFrame.
    It wraps the `pyspark.DataFrame.drop` function and can handle either a single string
    or a list of strings as input.

    If a specified column does not exist in the DataFrame, no error or warning is thrown,
    and all existing columns will remain.

    Expected behavior
    -----------------
    - When the `column` does not exist, all columns will remain (no error or warning is thrown)
    - Either a single string, or a list of strings can be specified

    Example
    -------
    __df__:

    |            product|amount|country|
    |-------------------|------|-------|
    |Banana lemon orange|  1000|    USA|
    |Carrots Blueberries|  1500|    USA|
    |              Beans|  1600|    USA|

    ```python
    output_df = DropColumn(column="product").transform(df)
    ```

    __output_df__:

    |amount|country|
    |------|-------|
    |  1000|    USA|
    |  1500|    USA|
    |  1600|    USA|

    In this example, the `product` column is dropped from the DataFrame `df`.
    """

    def execute(self):
        self.log.info(f"{self.column=}")
        self.output.df = self.df.drop(*self.columns)



from datetime import datetime

import pytest

from pyspark.sql import SparkSession
from pyspark.sql import functions as F

from koheesio.spark.transformations.row_number_dedup import RowNumberDedup

pytestmark = pytest.mark.spark


@pytest.mark.parametrize("target_column", ["col_row_nuber"])
def test_row_number_dedup(spark: SparkSession, target_column: str) -> None:
    df = spark.createDataFrame(
        [
            (
                "1",
                "a",
                "f1",
                datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f2",
                datetime.strptime("2023-10-29", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f3",
                datetime.strptime("2023-10-28", "%Y-%m-%d").date(),
            ),
        ],
        schema="key string, second_key string, field string, dt date",
    )
    assert (
        RowNumberDedup(df=df, columns=["key", "second_key"], sort_columns="dt", target_column=target_column)
        .transform()
        .head()
        .asDict()
    ) == {
        "key": "1",
        "second_key": "a",
        "field": "f1",
        "dt": datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
    }


@pytest.mark.parametrize("target_column", ["col_row_nuber"])
def test_row_number_dedup_not_list_column(spark: SparkSession, target_column: str) -> None:
    df = spark.createDataFrame(
        [
            (
                "1",
                "a",
                "f1",
                datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f2",
                datetime.strptime("2023-10-29", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f3",
                datetime.strptime("2023-10-28", "%Y-%m-%d").date(),
            ),
        ],
        schema="key string, second_key string, field string, dt date",
    )
    assert (
        RowNumberDedup(
            df=df, columns=["key", "second_key"], sort_columns=F.col("dt").desc(), target_column=target_column
        )
        .transform()
        .head()
        .asDict()
    ) == {
        "key": "1",
        "second_key": "a",
        "field": "f1",
        "dt": datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
    }


@pytest.mark.parametrize("target_column", ["col_row_nuber"])
def test_row_number_dedup_with_columns(spark: SparkSession, target_column: str) -> None:
    df = spark.createDataFrame(
        [
            (
                "1",
                "a",
                "f1",
                datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f2",
                datetime.strptime("2023-10-29", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f3",
                datetime.strptime("2023-10-28", "%Y-%m-%d").date(),
            ),
        ],
        schema="key string, second_key string, field string, dt date",
    )
    assert (
        RowNumberDedup(
            df=df, columns=["key", "second_key"], sort_columns=[F.col("dt").desc()], target_column=target_column
        )
        .transform()
        .head()
        .asDict()
    ) == {
        "key": "1",
        "second_key": "a",
        "field": "f1",
        "dt": datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
    }


@pytest.mark.parametrize("target_column", ["col_row_nuber"])
def test_row_number_dedup_with_duplicated_columns(spark: SparkSession, target_column: str) -> None:
    df = spark.createDataFrame(
        [
            (
                "1",
                "a",
                "f1",
                datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f2",
                datetime.strptime("2023-10-29", "%Y-%m-%d").date(),
            ),
            (
                "1",
                "a",
                "f3",
                datetime.strptime("2023-10-28", "%Y-%m-%d").date(),
            ),
        ],
        schema="key string, second_key string, field string, dt date",
    )
    transformation = RowNumberDedup(
        df=df,
        columns=["key", "second_key"],
        sort_columns=[F.col("dt").desc(), F.col("dt").desc(), F.col("dt").asc()],
        target_column=target_column,
    )

    assert [str(c) for c in transformation.sort_columns] == [str(c) for c in [F.col("dt").desc(), F.col("dt").asc()]]
    assert (transformation.transform().head().asDict()) == {
        "key": "1",
        "second_key": "a",
        "field": "f1",
        "dt": datetime.strptime("2023-10-30", "%Y-%m-%d").date(),
    }



"""
This module provides classes for asynchronous steps in the koheesio package.
"""

from typing import Dict, Union
from abc import ABC
from asyncio import iscoroutine

from koheesio.steps import Step, StepMetaClass, StepOutput


class AsyncStepMetaClass(StepMetaClass):
    """Metaclass for asynchronous steps.

    This metaclass is used to define asynchronous steps in the Koheesio framework.
    It inherits from the StepMetaClass and provides additional functionality for
    executing asynchronous steps.

    Attributes:
        None

    Methods:
        _execute_wrapper: Wrapper method for executing asynchronous steps.

    """

    def _execute_wrapper(cls, *args, **kwargs):
        """Wrapper method for executing asynchronous steps.

        This method is called when an asynchronous step is executed. It wraps the
        execution of the step with additional functionality.

        Parameters
        ----------
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns
        -------
            The result of executing the asynchronous step.

        """

        return super()._execute_wrapper(*args, **kwargs)


class AsyncStepOutput(Step.Output):
    """
    Represents the output of an asynchronous step.

    This class extends the base `Step.Output` class and provides additional functionality for merging key-value maps.

    Attributes
    ----------
    ...

    Methods
    -------
    merge(other: Union[Dict, StepOutput])
        Merge key-value map with self.
    """

    def merge(self, other: Union[Dict, StepOutput]):
        """Merge key,value map with self

        Examples
        --------
        ```python
        step_output = StepOutput(foo="bar")
        step_output.merge(
            {"lorem": "ipsum"}
        )  # step_output will now contain {'foo': 'bar', 'lorem': 'ipsum'}
        ```

        Functionally similar to adding two dicts together; like running `{**dict_a, **dict_b}`.

        Parameters
        ----------
        other: Union[Dict, StepOutput]
            Dict or another instance of a StepOutputs class that will be added to self
        """
        if isinstance(other, StepOutput):
            other = other.model_dump()  # ensures we really have a dict

        if not iscoroutine(other):
            for k, v in other.items():
                self.set(k, v)

        return self


class AsyncStep(Step, ABC, metaclass=AsyncStepMetaClass):
    """
    Asynchronous step class that inherits from Step and uses the AsyncStepMetaClass metaclass.

    Attributes
    ----------
    Output : AsyncStepOutput
        The output class for the asynchronous step.
    """

    class Output(AsyncStepOutput):
        """
        Output class for asyncio step.

        This class represents the output of the asyncio step. It inherits from the AsyncStepOutput class.
        """

    __output__: Output



"""Module that holds the transformations that can be used for date and time related operations."""

from typing import Optional, Union

from pytz import all_timezones_set

from pyspark.sql import Column
from pyspark.sql import functions as f
from pyspark.sql.functions import (
    col,
    date_format,
    from_utc_timestamp,
    lit,
    to_timestamp,
    to_utc_timestamp,
    when,
)

from koheesio.models import Field, field_validator, model_validator
from koheesio.spark.transformations import ColumnsTransformationWithTarget


def change_timezone(column: Union[str, Column], source_timezone: str, target_timezone: str):
    """Helper function to change from one timezone to another

    wrapper around `pyspark.sql.functions.from_utc_timestamp` and `to_utc_timestamp`

    Parameters
    ----------
    column : Union[str, Column]
        The column to change the timezone of
    source_timezone : str
        The timezone of the source_column value. Timezone fields are validated against the `TZ database name` column in
        this list: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    target_timezone : str
        The target timezone. Timezone fields are validated against the `TZ database name` column in this list:
        https://en.wikipedia.org/wiki/List_of_tz_database_time_zones

    """
    column = col(column) if isinstance(column, str) else column
    return from_utc_timestamp((to_utc_timestamp(column, source_timezone)), target_timezone)


class DateFormat(ColumnsTransformationWithTarget):
    """
    wrapper around pyspark.sql.functions.date_format

    See Also
    --------
    * https://spark.apache.org/docs/3.3.2/api/python/reference/pyspark.sql/api/pyspark.sql.functions.date_format.html
    * https://spark.apache.org/docs/3.3.2/sql-ref-datetime-pattern.html

    Concept
    -------
    This Transformation allows to convert a date/timestamp/string to a value of string in the format specified by the
    date format given.

    A pattern could be for instance `dd.MM.yyyy` and could return a string like ‘18.03.1993’.
    All pattern letters of datetime pattern can be used, see:
    https://spark.apache.org/docs/3.3.2/sql-ref-datetime-pattern.html

    How to use
    ----------
    If more than one column is passed, the behavior of the Class changes this way

    - the transformation will be run in a loop against all the given columns
    - the target_column will be used as a suffix. Leaving this blank will result in the original columns being renamed.

    Example
    -------
    ```python
    source_column value: datetime.date(2020, 1, 1)
    target: "yyyyMMdd HH:mm"
    output: "20200101 00:00"
    ```
    """

    format: str = Field(
        ...,
        description="The format for the resulting string. See "
        "https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html",
    )

    def func(self, column: Column) -> Column:
        return date_format(column, self.format)


class ChangeTimeZone(ColumnsTransformationWithTarget):
    """
    Allows for the value of a column to be changed from one timezone to another

    Adding useful metadata
    ----------------------
    When `add_target_timezone` is enabled (default), an additional column is created documenting which timezone a
    field has been converted to. Additionally, the suffix added to this column can be customized (default value is
    `_timezone`).

    Example
    -------

    Input:

    ```python
    target_column = "some_column_name"
    target_timezone = "EST"
    add_target_timezone = True  # default value
    timezone_column_suffix = "_timezone"  # default value
    ```

    Output:
    ```python
    column name  = "some_column_name_timezone"  # notice the suffix
    column value = "EST"
    ```

    <!--- TODO: add ability to change timezone depending on column values (future release) -->
    """

    from_timezone: str = Field(
        default=...,
        alias="source_timezone",
        description="Timezone of the source_column value. Timezone fields are validated against the `TZ database name` "
        "column in this list: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones",
    )
    to_timezone: str = Field(
        default=...,
        alias="target_timezone",
        description="Target timezone. Timezone fields are validated against the `TZ database name` column in this "
        "list: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones",
    )
    add_target_timezone: bool = Field(
        default=True, description="Toggles whether the target timezone is added as a column. True by default."
    )
    target_timezone_column_suffix: Optional[str] = Field(
        default="_timezone",
        alias="suffix",
        description="Allows to customize the suffix that is added to the target_timezone column. "
        "Defaults to '_timezone'. "
        "Note: this will be ignored if 'add_target_timezone' is set to False",
    )

    @model_validator(mode="before")
    def validate_no_duplicate_timezones(cls, values):
        """Validate that source and target timezone are not the same"""
        from_timezone_value = values.get("from_timezone")
        to_timezone_value = values.get("o_timezone")

        if from_timezone_value == to_timezone_value:
            raise ValueError("Timezone conversions from and to the same timezones are not valid.")

        return values

    @field_validator("from_timezone", "to_timezone")
    def validate_timezone(cls, timezone_value):
        """Validate that the timezone is a valid timezone."""
        if timezone_value not in all_timezones_set:
            raise ValueError(
                "Not a valid timezone. Refer to the `TZ database name` column here: "
                "https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
            )
        return timezone_value

    def func(self, column: Column) -> Column:
        return change_timezone(column=column, source_timezone=self.from_timezone, target_timezone=self.to_timezone)

    def execute(self):
        df = self.df

        for target_column, column in self.get_columns_with_target():
            func = self.func  # select the applicable function
            df = df.withColumn(
                target_column,
                func(f.col(column)),
            )

            # document which timezone a field has been converted to
            if self.add_target_timezone:
                df = df.withColumn(f"{target_column}{self.target_timezone_column_suffix}", f.lit(self.to_timezone))

        self.output.df = df


class ToTimestamp(ColumnsTransformationWithTarget):
    """
    wrapper around `pyspark.sql.functions.to_timestamp`

    Converts a Column (or set of Columns) into `pyspark.sql.types.TimestampType` using the specified format.
    Specify formats according to `datetime pattern https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html`_.

    Functionally equivalent to col.cast("timestamp").


    See Also
    --------
    Related Koheesio classes:

    [ColumnsTransformation]: ../index.md#koheesio.spark.transformations.ColumnsTransformation
    [ColumnsTransformationWithTarget]: ../index.md#koheesio.spark.transformations.ColumnsTransformationWithTarget
    [pyspark.sql.functions]: https://spark.apache.org/docs/3.5.1/api/python/reference/pyspark.sql/functions.html

    From the `koheesio.spark.transformations` module:

    * [ColumnsTransformation] :
        Base class for ColumnsTransformation. Defines column / columns field + recursive logic
    * [ColumnsTransformationWithTarget] :
        Defines target_column / target_suffix field

    [pyspark.sql.functions]:

    * datetime pattern : https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html

    Example
    -------

    ## Basic usage example:

    __input_df:__

    | t                     |
    |-----------------------|
    | "1997-02-28 10:30:00" |

    `t` is a string

    ```python
    tts = ToTimestamp(
        # since the source column is the same as the target in this example, 't' will be overwritten
        column="t",
        target_column="t",
        format="yyyy-MM-dd HH:mm:ss",
    )
    output_df = tts.transform(input_df)
    ```

    __output_df:__

    | t                                      |
    |----------------------------------------|
    | datetime.datetime(1997, 2, 28, 10, 30) |

    Now `t` is a timestamp


    ## Multiple columns at once:

    __input_df:__

    | t1                    | t2                     |
    |-----------------------|------------------------|
    | "1997-02-28 10:30:00" | "2007-03-31 11:40:10"  |

    `t1` and `t2` are strings

    ```python
    tts = ToTimestamp(
        columns=["t1", "t2"],
        # 'target_suffix' is synonymous with 'target_column'
        target_suffix="new",
        format="yyyy-MM-dd HH:mm:ss",
    )
    output_df = tts.transform(input_df).select("t1_new", "t2_new")
    ```

    __output_df:__

    | t1_new                                 | t2_new                                  |
    |----------------------------------------|-----------------------------------------|
    | datetime.datetime(1997, 2, 28, 10, 30) | datetime.datetime(2007, 3, 31, 11, 40)  |

    Now `t1_new` and `t2_new` are both timestamps

    """

    format: str = Field(
        default=...,  # TODO: should we set a logic default here? [future release]
        description="The date format for of the timestamp field. See "
        "https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html",
    )

    def func(self, column: Column) -> Column:
        # convert string to timestamp
        converted_col = to_timestamp(column, self.format)
        return when(column.isNull(), lit(None)).otherwise(converted_col)



"""Models package creates models that can be used to base other classes on.

- Every model should be at least a pydantic BaseModel, but can also be a Step, or a StepOutput.
- Every model is expected to be an ABC (Abstract Base Class)
- Optionally a model can inherit ExtraParamsMixin that provides unpacking of kwargs into `extra_params` dict property
  removing need to create a dict before passing kwargs to a model initializer.

A Model class can be exceptionally handy when you need similar Pydantic models in multiple places, for example across
Transformation and Reader classes.
"""

from typing import Annotated, Any, Dict, List, Optional, Union
from abc import ABC
from functools import cached_property
from pathlib import Path

# to ensure that koheesio.models is a drop in replacement for pydantic
from pydantic import BaseModel as PydanticBaseModel
from pydantic import *  # noqa
from pydantic._internal._generics import PydanticGenericMetadata
from pydantic._internal._model_construction import ModelMetaclass

from koheesio.context import Context
from koheesio.logger import Logger, LoggingFactory

__all__ = [
    "BaseModel",
    "ExtraParamsMixin",
    "Field",
    "ListOfColumns",
    "ModelMetaclass",
    "PydanticGenericMetadata",
    "field_validator",
    "model_validator",
]


# pylint: disable=function-redefined
class BaseModel(PydanticBaseModel, ABC):
    """
    Base model for all models.

    Extends pydantic BaseModel with some additional configuration.
    To be used as a base class for all models in Koheesio instead of pydantic.BaseModel.

    Additional methods and properties:
    ---------------------------------
    ### Fields
    Every Koheesio BaseModel has two predefined fields: `name` and `description`. These fields are used to provide a
    name and a description to the model.

    - `name`: This is the name of the Model. If not provided, it defaults to the class name.

    - `description`: This is the description of the Model. It has several default behaviors:
        - If not provided, it defaults to the docstring of the class.
        - If the docstring is not provided, it defaults to the name of the class.
        - For multi-line descriptions, it has the following behaviors:
            - Only the first non-empty line is used.
            - Empty lines are removed.
            - Only the first 3 lines are considered.
            - Only the first 120 characters are considered.

    ### Validators
    - `_set_name_and_description`: Set the name and description of the Model as per the rules mentioned above.

    ### Properties
    - `log`: Returns a logger with the name of the class.

    ### Class Methods
    - `from_basemodel`: Returns a new BaseModel instance based on the data of another BaseModel.
    - `from_context`: Creates BaseModel instance from a given Context.
    - `from_dict`: Creates BaseModel instance from a given dictionary.
    - `from_json`: Creates BaseModel instance from a given JSON string.
    - `from_toml`: Creates BaseModel object from a given toml file.
    - `from_yaml`: Creates BaseModel object from a given yaml file.
    - `lazy`: Constructs the model without doing validation.

    ### Dunder Methods
    - `__add__`: Allows to add two BaseModel instances together.
    - `__enter__`: Allows for using the model in a with-statement.
    - `__exit__`: Allows for using the model in a with-statement.
    - `__setitem__`: Set Item dunder method for BaseModel.
    - `__getitem__`: Get Item dunder method for BaseModel.

    ### Instance Methods
    - `hasattr`: Check if given key is present in the model.
    - `get`: Get an attribute of the model, but don't fail if not present.
    - `merge`: Merge key,value map with self.
    - `set`: Allows for subscribing / assigning to `class[key]`.
    - `to_context`: Converts the BaseModel instance to a Context object.
    - `to_dict`: Converts the BaseModel instance to a dictionary.
    - `to_json`: Converts the BaseModel instance to a JSON string.
    - `to_yaml`: Converts the BaseModel instance to a YAML string.

    Different Modes
    ---------------
    This BaseModel class supports lazy mode. This means that validation of the items stored in the class can be called
    at will instead of being forced to run it upfront.

    * _Normal mode_:
        you need to know the values ahead of time
        ```python
        normal_mode = YourOwnModel(a="foo", b=42)
        ```

    * _Lazy mode_:
        being able to defer the validation until later
        ```python
        lazy_mode = YourOwnModel.lazy()
        lazy_mode.a = "foo"
        lazy_mode.b = 42
        lazy_mode.validate_output()
        ```
        The prime advantage of using lazy mode is that you don't have to know all your outputs up front, and can add
        them as they become available. All while still being able to validate that you have collected all your output
        at the end.

    * _With statements_:
        With statements are also allowed. The `validate_output` method from the earlier example will run upon exit of
        the with-statement.
        ```python
        with YourOwnModel.lazy() as with_output:
            with_output.a = "foo"
            with_output.b = 42
        ```
        Note: that a lazy mode BaseModel object is required to work with a with-statement.

    Examples
    --------
    ```python
    from koheesio.models import BaseModel


    class Person(BaseModel):
        name: str
        age: int


    # Using the lazy method to create an instance without immediate validation
    person = Person.lazy()

    # Setting attributes
    person.name = "John Doe"
    person.age = 30

    # Now we validate the instance
    person.validate_output()

    print(person)
    ```

    In this example, the Person instance is created without immediate validation. The attributes name and age are set
    afterward. The `validate_output` method is then called to validate the instance.

    Koheesio specific configuration:
    -------------------------------
    Koheesio models are configured differently from Pydantic defaults. The following configuration is used:

    1. *extra="allow"*\n
        This setting allows for extra fields that are not specified in the model definition. If a field is present in
        the data but not in the model, it will not raise an error.
        Pydantic default is "ignore", which means that extra attributes are ignored.

    2. *arbitrary_types_allowed=True*\n
        This setting allows for fields in the model to be of any type. This is useful when you want to include fields
        in your model that are not standard Python types.
        Pydantic default is False, which means that fields must be of a standard Python type.

    3. *populate_by_name=True*\n
        This setting allows an aliased field to be populated by its name as given by the model attribute, as well as
        the alias. This was known as allow_population_by_field_name in pydantic v1.
        Pydantic default is False, which means that fields can only be populated by their alias.

    4. *validate_assignment=False*\n
        This setting determines whether the model should be revalidated when the data is changed. If set to `True`,
        every time a field is assigned a new value, the entire model is validated again.\n
        Pydantic default is (also) `False`, which means that the model is not revalidated when the data is changed.
        The default behavior of Pydantic is to validate the data when the model is created. In case the user changes
        the data after the model is created, the model is _not_ revalidated.

    5. *revalidate_instances="subclass-instances"*\n
        This setting determines whether to revalidate models during validation if the instance is a subclass of the
        model. This is important as inheritance is used a lot in Koheesio.
        Pydantic default is `never`, which means that the model and dataclass instances are not revalidated during
        validation.

    6. *validate_default=True*\n
        This setting determines whether to validate default values during validation. When set to True, default values
        are checked during the validation process. We opt to set this to True, as we are attempting to make the sure
        that the data is valid prior to running / executing any Step.
        Pydantic default is False, which means that default values are not validated during validation.

    7. *frozen=False*\n
        This setting determines whether the model is immutable. If set to True, once a model is created, its fields
        cannot be changed.
        Pydantic default is also False, which means that the model is mutable.

    8. *coerce_numbers_to_str=True*\n
        This setting determines whether to convert number fields to strings. When set to True, enables automatic
        coercion of any `Number` type to `str`.
        Pydantic doesn't allow number types (`int`, `float`, `Decimal`) to be coerced as type `str` by default.

    9. *use_enum_values=True*\n
        This setting determines whether to use the values of Enum fields. If set to True, the actual value of the Enum
        is used instead of the reference.
        Pydantic default is False, which means that the reference to the Enum is used.
    """

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        populate_by_name=True,
        validate_assignment=False,
        revalidate_instances="subclass-instances",
        validate_default=True,
        frozen=False,
        coerce_numbers_to_str=True,
        use_enum_values=True,
    )

    name: Optional[str] = Field(default=None, description="Name of the Model")
    description: Optional[str] = Field(default=None, description="Description of the Model")

    @model_validator(mode="after")
    def _validate_name_and_description(self):
        """
        Validates the 'name' and 'description' of the Model according to the rules outlined in the class docstring.
        """
        self.name = str(self.name or self.__class__.__name__ or "")
        _description = (self.description or self.__doc__ or self.name).split("\n", maxsplit=2)
        self.description = next((line for line in _description if line.strip()), "").strip()

        # Limit the description to around 120 characters, cutting at the first whole word exceeding this limit
        if len(self.description) > 120:
            # Find the first space after the 115th character
            if (space_index := self.description.find(" ", 115)) == -1:
                space_index = 117  # 120 characters - 3 for the ellipsis

            self.description = self.description[:space_index] + "..."

        return self

    @property
    def log(self) -> Logger:
        """Returns a logger with the name of the class"""
        return LoggingFactory.get_logger(name=self.__class__.__name__, inherit_from_koheesio=True)

    @classmethod
    def from_basemodel(cls, basemodel: BaseModel, **kwargs) -> InstanceOf[BaseModel]:
        """Returns a new BaseModel instance based on the data of another BaseModel"""
        kwargs = {**basemodel.model_dump(), **kwargs}
        return cls(**kwargs)

    @classmethod
    def from_context(cls, context: Context) -> BaseModel:
        """Creates BaseModel instance from a given Context

        You have to make sure that the Context object has the necessary attributes to create the model.

        Examples
        --------
        ```python
        class SomeStep(BaseModel):
            foo: str


        context = Context(foo="bar")
        some_step = SomeStep.from_context(context)
        print(some_step.foo)  # prints 'bar'
        ```

        Parameters
        ----------
        context: Context

        Returns
        -------
        BaseModel
        """
        return cls(**context)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> BaseModel:
        """Creates BaseModel instance from a given dictionary

        Parameters
        ----------
        data: Dict[str, Any]

        Returns
        -------
        BaseModel
        """
        return cls(**data)

    @classmethod
    def from_json(cls, json_file_or_str: Union[str, Path]) -> BaseModel:
        """Creates BaseModel instance from a given JSON string

        BaseModel offloads the serialization and deserialization of the JSON string to Context class. Context uses
        jsonpickle library to serialize and deserialize the JSON string. This is done to allow for objects to be stored
        in the BaseModel object, which is not possible with the standard json library.

        See Also
        --------
        Context.from_json : Deserializes a JSON string to a Context object

        Parameters
        ----------
        json_file_or_str : Union[str, Path]
            Pathlike string or Path that points to the json file or string containing json

        Returns
        -------
        BaseModel
        """
        _context = Context.from_json(json_file_or_str)
        return cls.from_context(_context)

    @classmethod
    def from_toml(cls, toml_file_or_str: Union[str, Path]) -> BaseModel:
        """Creates BaseModel object from a given toml file

        Note: BaseModel offloads the serialization and deserialization of the TOML string to Context class.

        Parameters
        ----------
        toml_file_or_str: str or Path
            Pathlike string or Path that points to the toml file, or string containing toml

        Returns
        -------
        BaseModel
        """
        _context = Context.from_toml(toml_file_or_str)
        return cls.from_context(_context)

    @classmethod
    def from_yaml(cls, yaml_file_or_str: str) -> BaseModel:
        """Creates BaseModel object from a given yaml file

        Note: BaseModel offloads the serialization and deserialization of the YAML string to Context class.

        Parameters
        ----------
        yaml_file_or_str: str or Path
            Pathlike string or Path that points to the yaml file, or string containing yaml

        Returns
        -------
        BaseModel
        """
        _context = Context.from_yaml(yaml_file_or_str)
        return cls.from_context(_context)

    @classmethod
    def lazy(cls):
        """Constructs the model without doing validation

        Essentially an alias to BaseModel.construct()
        """
        return cls.model_construct()

    def __add__(self, other: Union[Dict, BaseModel]) -> BaseModel:
        """Allows to add two BaseModel instances together

        Essentially a shorthand for running .merge against a second BaseModel object

        Examples
        --------
        ```python
        step_output_1 = StepOutput(foo="bar")
        step_output_2 = StepOutput(lorem="ipsum")
        (
            step_output_1 + step_output_2
        )  # step_output_1 will now contain {'foo': 'bar', 'lorem': 'ipsum'}
        ```

        Parameters
        ----------
        other: Union[Dict, BaseModel]
            Dict or another instance of a BaseModel class that will be added to self

        Returns
        -------
        None
            Add is additive and modifies the object in place, hence there is no return value
        """
        return self.merge(other)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            # An exception occurred. We log it and raise it again.
            self.log.exception(f"An exception occurred: {exc_val}")
            return False  # this will re-raise the exception

        # No exception occurred. We validate the final state of the BaseModel.
        self.validate()
        return True

    def __getitem__(self, name) -> Any:
        """Get Item dunder method for BaseModel

        Allows for subscriptable (`class[key]`) type of access to the data.

        Examples
        --------
        ```python
        step_output = StepOutput(foo="bar")
        step_output["foo"]  # returns 'bar'
        ```

        Parameters
        ----------
        name: str
            The name of the attribute

        Returns
        -------
        Any
            The value of the attribute
        """
        return self.__getattribute__(name)

    def __setitem__(self, key: str, value: Any):
        """Set Item dunder method for BaseModel

        Allows for subscribing / assigning to `class[key]`

        Examples
        --------
        ```python
        step_output = StepOutput(foo="bar")
        step_output["foo"] = "baz"  # overwrites 'foo' to be 'baz'
        ```

        Parameters
        ----------
        key: str
            The key of the attribute to assign to
        value: Any
            Value that should be assigned to the given key
        """
        self.__setattr__(key, value)

    def hasattr(self, key: str) -> bool:
        """Check if given key is present in the model

        Parameters
        ----------
        key: str

        Returns
        -------
        bool
        """
        return hasattr(self, key)

    def get(self, key: str, default: Optional[Any] = None):
        """Get an attribute of the model, but don't fail if not present

        Similar to dict.get()

        Examples
        --------
        ```python
        step_output = StepOutput(foo="bar")
        step_output.get("foo")  # returns 'bar'
        step_output.get("non_existent_key", "oops")  # returns 'oops'
        ```

        Parameters
        ----------
        key: str
            name of the key to get
        default: Optional[Any]
            Default value in case the attribute does not exist

        Returns
        -------
        Any
            The value of the attribute
        """
        if self.hasattr(key):
            return self.__getitem__(key)
        return default

    def merge(self, other: Union[Dict, BaseModel]):
        """Merge key,value map with self

        Functionally similar to adding two dicts together; like running `{**dict_a, **dict_b}`.

        Examples
        --------
        ```python
        step_output = StepOutput(foo="bar")
        step_output.merge(
            {"lorem": "ipsum"}
        )  # step_output will now contain {'foo': 'bar', 'lorem': 'ipsum'}
        ```

        Parameters
        ----------
        other: Union[Dict, BaseModel]
            Dict or another instance of a BaseModel class that will be added to self
        """
        if isinstance(other, BaseModel):
            other = other.model_dump()  # ensures we really have a dict

        for k, v in other.items():
            self.set(k, v)

        return self

    def set(self, key: str, value: Any):
        """Allows for subscribing / assigning to `class[key]`.

        Examples
        --------
        ```python
        step_output = StepOutput(foo="bar")
        step_output.set(foo", "baz")  # overwrites 'foo' to be 'baz'
        ```

        Parameters
        ----------
        key: str
            The key of the attribute to assign to
        value: Any
            Value that should be assigned to the given key
        """
        self.__setitem__(key, value)

    def to_context(self) -> Context:
        """Converts the BaseModel instance to a Context object

        Returns
        -------
        Context
        """
        return Context(**self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        """Converts the BaseModel instance to a dictionary

        Returns
        -------
        Dict[str, Any]
        """
        return self.model_dump()

    def to_json(self, pretty: bool = False):
        """Converts the BaseModel instance to a JSON string

        BaseModel offloads the serialization and deserialization of the JSON string to Context class. Context uses
        jsonpickle library to serialize and deserialize the JSON string. This is done to allow for objects to be stored
        in the BaseModel object, which is not possible with the standard json library.

        See Also
        --------
        Context.to_json : Serializes a Context object to a JSON string

        Parameters
        ----------
        pretty : bool, optional, default=False
            Toggles whether to return a pretty json string or not

        Returns
        -------
        str
            containing all parameters of the BaseModel instance
        """
        _context = self.to_context()
        return _context.to_json(pretty=pretty)

    def to_yaml(self, clean: bool = False) -> str:
        """Converts the BaseModel instance to a YAML string

        BaseModel offloads the serialization and deserialization of the YAML string to Context class.

        Parameters
        ----------
        clean: bool
            Toggles whether to remove `!!python/object:...` from yaml or not.
            Default: False

        Returns
        -------
        str
            containing all parameters of the BaseModel instance
        """
        _context = self.to_context()
        return _context.to_yaml(clean=clean)

    # noinspection PyMethodOverriding
    def validate(self) -> BaseModel:
        """Validate the BaseModel instance

        This method is used to validate the BaseModel instance. It is used in conjunction with the lazy method to
        validate the instance after all the attributes have been set.

        This method is intended to be used with the `lazy` method. The `lazy` method is used to create an instance of
        the BaseModel without immediate validation. The `validate` method is then used to validate the instance after.

        > Note: in the Pydantic BaseModel, the `validate` method throws a deprecated warning. This is because Pydantic
        recommends using the `validate_model` method instead. However, we are using the `validate` method here in a
        different context and a slightly different way.

        Examples
        --------
        ```python
        class FooModel(BaseModel):
            foo: str
            lorem: str


        foo_model = FooModel.lazy()
        foo_model.foo = "bar"
        foo_model.lorem = "ipsum"
        foo_model.validate()
        ```
        In this example, the `foo_model` instance is created without immediate validation. The attributes foo and lorem
        are set afterward. The `validate` method is then called to validate the instance.

        Returns
        -------
        BaseModel
            The BaseModel instance
        """
        return self.model_validate(self.model_dump())


# pylint: enable=function-redefined


class ExtraParamsMixin(PydanticBaseModel):
    """
    Mixin class that adds support for arbitrary keyword arguments to Pydantic models.

    The keyword arguments are extracted from the model's `values` and moved to a `params` dictionary.
    """

    params: Dict[str, Any] = Field(default_factory=dict)

    @cached_property
    def extra_params(self) -> Dict[str, Any]:
        """Extract params (passed as arbitrary kwargs) from values and move them to params dict"""
        # noinspection PyUnresolvedReferences
        return self.model_extra

    @model_validator(mode="after")
    def _move_extra_params_to_params(self):
        """Move extra_params to params dict"""
        self.params = {**self.params, **self.extra_params}
        return self


def _list_of_columns_validation(columns_value):
    """
    Performs validation for ListOfColumns type. Will ensure that there are no duplicate columns, empty strings, etc.
    In case an individual column is passed, it will coerce it to a list.
    """
    columns = [columns_value] if isinstance(columns_value, str) else [*columns_value]
    columns = [col for col in columns if col]  # remove empty strings, None, etc.

    return list(dict.fromkeys(columns))  # dict.fromkeys is used to dedup while maintaining order


ListOfColumns = Annotated[Union[str, List[str]], BeforeValidator(_list_of_columns_validation)]
""" Annotated type for a list of column names. 
Will ensure that there are no duplicate columns, empty strings, etc.
In case an individual column is passed, the value will be coerced to a list.
"""



"""
Classes to ease interaction with Slack
"""

import json
from typing import Any, Dict, Optional
from datetime import datetime
from textwrap import dedent

from koheesio.models import ConfigDict, Field
from koheesio.notifications import NotificationSeverity
from koheesio.steps.http import HttpPostStep


class SlackNotification(HttpPostStep):
    """
    Generic Slack notification class via the `Blocks` API

    > NOTE: `channel` parameter is used only with Slack Web API: https://api.slack.com/messaging/sending
        If webhook is used, the channel specification is not required

    Example:
    ```python
    s = SlackNotification(
        url="slack-webhook-url",
        channel="channel",
        message="Some *markdown* compatible text",
    )
    s.execute()
    ```
    """

    message: str = Field(default=..., description="The message that gets posted to Slack")
    channel: Optional[str] = Field(default=None, description="Slack channel id")
    headers: Optional[Dict[str, Any]] = {"Content-type": "application/json"}

    def get_payload(self):
        """
        Generate payload with `Block Kit`.
        More details: https://api.slack.com/block-kit
        """
        payload = {
            "attachments": [
                {
                    "blocks": [
                        {
                            "type": "section",
                            "text": {
                                "type": "mrkdwn",
                                "text": self.message,
                            },
                        }
                    ],
                }
            ]
        }

        if self.channel:
            payload["channel"] = self.channel

        return json.dumps(payload)

    def execute(self):
        """
        Generate payload and send post request
        """
        self.data = self.get_payload()
        HttpPostStep.execute(self)


class SlackNotificationWithSeverity(SlackNotification):
    """
    Slack notification class via the `Blocks` API with etra severity information and predefined extra fields

    Example:
        from koheesio.steps.integrations.notifications import NotificationSeverity

        s = SlackNotificationWithSeverity(
            url="slack-webhook-url",
            channel="channel",
            message="Some *markdown* compatible text"
            severity=NotificationSeverity.ERROR,
            title="Title",
            environment="dev",
            application="Application"
        )
        s.execute()
    """

    severity: NotificationSeverity = Field(default=..., description="Severity of the message")
    title: str = Field(default=..., description="Title of your message")
    environment: str = Field(default=..., description="Environment description, e.g. dev / qa /prod")
    application: str = Field(default=..., description="Pipeline or application name")
    timestamp: datetime = Field(
        default=datetime.utcnow(),
        alias="execution_timestamp",
        description="Pipeline or application execution timestamp",
    )

    model_config = ConfigDict(use_enum_values=False)

    def get_payload_message(self):
        """
        Generate payload message based on the predefined set of parameters
        """
        return dedent(
            f"""
                {self.severity.alert_icon}   *{self.severity.name}:*  {self.title}
                *Environment:* {self.environment}
                *Application:* {self.application}
                *Message:* {self.message}
                *Timestamp:* {self.timestamp}
            """
        )

    def execute(self):
        """
        Generate payload and send post request
        """
        self.message = self.get_payload_message()
        self.data = self.get_payload()
        HttpPostStep.execute(self)



"""
ETL Task

Extract -> Transform -> Load
"""

from datetime import datetime

from pyspark.sql import DataFrame

from koheesio import Step
from koheesio.models import Field, InstanceOf, conlist
from koheesio.spark.readers import Reader
from koheesio.spark.transformations import Transformation
from koheesio.spark.writers import Writer


class EtlTask(Step):
    """ETL Task

    Etl stands for: Extract -> Transform -> Load

    This task is a composition of a Reader (extract), a series of Transformations (transform) and a Writer (load).
    In other words, it _reads_ data from a _source_, applies a series of _transformations_, and _writes_ the result to
    a _target_.

    Parameters
    ----------
    name : str
        Name of the task
    description : str
        Description of the task
    source : koheesio.steps.readers.Reader
        Source to read from [extract]
    transformations : list[koheesio.steps.transformations.Transformation]
        Series of transformations [transform]. The order of the transformations is important!
    target : koheesio.steps.writers.Writer
        Target to write to [load]


    Example
    -------
    ```python
    from koheesio.tasks import EtlTask

    from koheesio.steps.readers import CsvReader
    from koheesio.steps.transformations.repartition import Repartition
    from koheesio.steps.writers import CsvWriter

    etl_task = EtlTask(
        name="My ETL Task",
        description="This is an example ETL task",
        source=CsvReader(path="path/to/source.csv"),
        transformations=[Repartition(num_partitions=2)],
        target=DummyWriter(),
    )

    etl_task.execute()
    ```

    This code will read from a CSV file, repartition the DataFrame to 2 partitions, and write the result to the console.

    Extending the EtlTask
    ---------------------
    The EtlTask is designed to be a simple and flexible way to define ETL processes. It is not designed to be a
    one-size-fits-all solution, but rather a starting point for building more complex ETL processes. If you need more
    complex functionality, you can extend the EtlTask class and override the `extract`, `transform` and `load` methods.
    You can also implement your own `execute` method to define the entire ETL process from scratch should you need more
    flexibility.

    Advantages of using the EtlTask
    -------------------------------
    - It is a simple way to define ETL processes
    - It is easy to understand and extend
    - It is easy to test and debug
    - It is easy to maintain and refactor
    - It is easy to integrate with other tools and libraries
    - It is easy to use in a production environment
    """

    source: InstanceOf[Reader] = Field(default=..., description="Source to read from [extract]")
    transformations: conlist(min_length=0, item_type=InstanceOf[Transformation]) = Field(
        default_factory=list, description="Series of transformations", alias="transforms"
    )
    target: InstanceOf[Writer] = Field(default=..., description="Target to write to [load]")

    # private attrs
    etl_date: datetime = Field(
        default=datetime.utcnow(),
        description="Date time when this object was created as iso format. Example: '2023-01-24T09:39:23.632374'",
    )

    class Output(Step.Output):
        """Output class for EtlTask"""

        source_df: DataFrame = Field(default=..., description="The Spark DataFrame produced by .extract() method")
        transform_df: DataFrame = Field(default=..., description="The Spark DataFrame produced by .transform() method")
        target_df: DataFrame = Field(default=..., description="The Spark DataFrame used by .load() method")

    def extract(self) -> DataFrame:
        """Read from Source

        logging is handled by the Reader.execute()-method's @do_execute decorator
        """
        reader: Reader = self.source
        return reader.read()

    def transform(self, df: DataFrame) -> DataFrame:
        """Transform recursively

        logging is handled by the Transformation.execute()-method's @do_execute decorator
        """
        for t in self.transformations:
            df = t.transform(df)
        return df

    def load(self, df: DataFrame) -> DataFrame:
        """Write to Target

        logging is handled by the Writer.execute()-method's @do_execute decorator
        """
        writer: Writer = self.target
        writer.write(df)
        return df

    def execute(self):
        """Run the ETL process"""
        self.log.info(f"Task started at {self.etl_date}")

        # extract from source
        self.output.source_df = self.extract()

        # transform
        self.output.transform_df = self.transform(self.output.source_df)

        # load to target
        self.output.target_df = self.load(self.output.transform_df)

    def run(self):
        """alias of execute"""
        return self.execute()



"""
A simple DummyReader that returns a DataFrame with an id-column of the given range
"""

from koheesio.models import Field
from koheesio.spark.readers import Reader


class DummyReader(Reader):
    """A simple DummyReader that returns a DataFrame with an id-column of the given range

    Can be used in place of any Reader without having to read from a real source.

    Wraps SparkSession.range(). Output DataFrame will have a single column named "id" of type Long and length of the
    given range.

    Parameters
    ----------
    range : int
        How large to make the Dataframe

    Example
    -------
    ```python
    from koheesio.spark.readers.dummy import DummyReader

    output_df = DummyReader(range=100).read()
    ```

    __output_df__:
    Output DataFrame will have a single column named "id" of type `Long` containing 100 rows (0-99).

    | id  |
    |-----|
    | 0   |
    | 1   |
    | ... |
    | 99  |
    """

    range: int = Field(default=100, description="How large to make the Dataframe")

    def execute(self):
        self.output.df = self.spark.range(self.range)



import logging
from io import StringIO
from logging import Logger
from unittest.mock import MagicMock, patch

import pytest

from koheesio.logger import (
    LoggingFactory,
    MaskedDict,
    MaskedFloat,
    MaskedInt,
    MaskedString,
)


class TestLoggingFactory:
    def test_create_logger(self):
        logger = LoggingFactory.get_logger(name="TestLogger")
        assert isinstance(logger, Logger)

    def test_get_logger(self):
        child_name = "child_logger"
        ind_name = "independent_logger"
        logger = LoggingFactory.get_logger(name=child_name, inherit_from_koheesio=True)
        logger.debug("Test Child DEBUG")
        logger_independent = LoggingFactory.get_logger(name=ind_name)
        logger_independent.setLevel("INFO")

        assert isinstance(logger, Logger)
        assert logger.name == f"{LoggingFactory.LOGGER_NAME}.{child_name}"
        assert logger.parent.level == logging._nameToLevel[LoggingFactory.LOGGER_LEVEL]
        assert isinstance(logger_independent, Logger)
        assert logger_independent.name == ind_name
        assert logger_independent.level == logging.INFO


class TestAddHandlers:
    @pytest.fixture
    def mock_import_class(self):
        with patch("koheesio.logger.import_class") as mock:
            yield mock

    @pytest.fixture
    def mock_handler_class(self, mock_import_class):
        mock_handler_class = MagicMock()
        mock_import_class.return_value = mock_handler_class
        return mock_handler_class

    @pytest.fixture
    def mock_handler_instance(self, mock_handler_class):
        mock_handler_instance = MagicMock()
        mock_handler_class.return_value = mock_handler_instance
        return mock_handler_instance

    @pytest.fixture
    def handlers_config(self):
        return [("handler_module.Class1", {"level": "DEBUG"})]

    def test_add_handlers(self, mock_import_class, mock_handler_instance, handlers_config):
        # Configure the LoggingFactory attributes
        LoggingFactory.LOGGER = MagicMock()
        LoggingFactory.LOGGER_FILTER = MagicMock()
        LoggingFactory.LOGGER_FORMATTER = MagicMock()

        # Call the add_handlers method
        LoggingFactory.add_handlers(handlers_config)

        # Assertions
        mock_import_class.assert_called_with("handler_module.Class1")
        mock_handler_instance.setLevel.assert_called_with("DEBUG")
        mock_handler_instance.addFilter.assert_called_with(LoggingFactory.LOGGER_FILTER)
        mock_handler_instance.setFormatter.assert_called_with(LoggingFactory.LOGGER_FORMATTER)
        LoggingFactory.LOGGER.addHandler.assert_called_with(mock_handler_instance)


@pytest.mark.parametrize(
    "foo, bar, baz, params",
    [
        pytest.param(
            MaskedString("some value"),
            MaskedInt(123),
            MaskedFloat(1.23),
            MaskedDict({"key": "value"}),
            id="test_log_masked",
        ),
    ],
)
def test_log_masked(foo, bar, baz, params):
    log_capture_string = StringIO()
    ch = logging.StreamHandler(log_capture_string)
    ch.setLevel(logging.DEBUG)
    logger = logging.getLogger("tests")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(ch)

    logger.warning(f"{foo = }")
    logger.warning(foo)

    log_contents = log_capture_string.getvalue().split("\n")
    assert log_contents[-3] == "foo = '**********(Masked)'"
    assert log_contents[-2] == "'**********(Masked)'"

    logger.warning(f"{bar = }")
    logger.warning("this is supposed to be a masked field: %s", baz)
    log_contents = log_capture_string.getvalue().split("\n")
    assert log_contents[-3] == "bar = '***(Masked)'"
    assert log_contents[-2] == "this is supposed to be a masked field: '****(Masked)'"

    logger.warning(MaskedFloat(bar + baz))
    log_contents = log_capture_string.getvalue().split("\n")
    assert log_contents[-2] == "'******(Masked)'"

    logger.info("this is supposed to be a masked field2: %s", params)
    log_contents = log_capture_string.getvalue().split("\n")
    print(log_contents)
    assert log_contents[-2] == "this is supposed to be a masked field2: '****************(Masked)'"
